import warnings
from typing import Any, Callable, List, Optional, Union

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...utils import load_torch_weights, MODEL_URLS


class GoogLeNet(eqx.Module):
    """A simple port of `torchvision.models.GoogLeNet`"""

    aux_logits: bool
    conv1: eqx.Module
    maxpool1: nn.MaxPool2d
    conv2: eqx.Module
    conv3: eqx.Module
    maxpool2: nn.MaxPool2d
    inception3a: eqx.Module
    inception3b: eqx.Module
    maxpool3: nn.MaxPool2d
    inception4a: eqx.Module
    inception4b: eqx.Module
    inception4c: eqx.Module
    inception4d: eqx.Module
    inception4e: eqx.Module
    maxpool4: nn.MaxPool2d
    inception5a: eqx.Module
    inception5b: eqx.Module
    aux1: eqx.Module
    aux2: eqx.Module
    avgpool: nn.AdaptiveAvgPool2d
    dropout: nn.Dropout
    fc: nn.Linear

    def __init__(
        self,
        num_classes: int = 1000,
        aux_logits: bool = False,
        blocks: Optional[List["eqx.Module"]] = None,
        dropout: float = 0.2,
        dropout_aux: float = 0.7,
        *,
        key: Optional["jax.random.PRNGKey"] = None
    ) -> None:
        """
        **Arguments:**

        - `num_classes`: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        - `aux_logits`: If `True`, two auxiliary branches are added to the network. Defaults to `False`
        - `blocks`: Blocks for constructing the network
        - `dropout`: Dropout applied on the `main` branch. Defaults to `0.2`
        - `dropout_aux`: Dropout applied on the `aux` branches. Defaults to `0.7`
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)

        """
        super().__init__()
        if blocks is None:
            blocks = [BasicConv2d, _Inception, InceptionAux]
        assert len(blocks) == 3
        conv_block = blocks[0]
        inception_block = blocks[1]
        inception_aux_block = blocks[2]

        if key is None:
            key = jrandom.PRNGKey(0)
        keys = jrandom.split(key, 20)

        self.aux_logits = aux_logits
        self.conv1 = conv_block(3, 64, kernel_size=7, stride=2, padding=3, key=keys[0])
        self.maxpool1 = nn.MaxPool2d(3, stride=2, use_ceil=True)
        self.conv2 = conv_block(64, 64, kernel_size=1, key=keys[1])
        self.conv3 = conv_block(64, 192, kernel_size=3, padding=1, key=keys[2])
        self.maxpool2 = nn.MaxPool2d(3, stride=2, use_ceil=True)

        self.inception3a = inception_block(192, 64, 96, 128, 16, 32, 32, key=keys[3])
        self.inception3b = inception_block(256, 128, 128, 192, 32, 96, 64, key=keys[4])
        self.maxpool3 = nn.MaxPool2d(3, stride=2, use_ceil=True)

        self.inception4a = inception_block(480, 192, 96, 208, 16, 48, 64, key=keys[5])
        self.inception4b = inception_block(512, 160, 112, 224, 24, 64, 64, key=keys[6])
        self.inception4c = inception_block(512, 128, 128, 256, 24, 64, 64, key=keys[7])
        self.inception4d = inception_block(512, 112, 144, 288, 32, 64, 64, key=keys[8])
        self.inception4e = inception_block(
            528, 256, 160, 320, 32, 128, 128, key=keys[9]
        )
        self.maxpool4 = nn.MaxPool2d(2, stride=2, use_ceil=True)

        self.inception5a = inception_block(
            832, 256, 160, 320, 32, 128, 128, key=keys[10]
        )
        self.inception5b = inception_block(
            832, 384, 192, 384, 48, 128, 128, key=keys[11]
        )

        self.aux1 = None
        self.aux2 = None
        if aux_logits:
            self.aux1 = inception_aux_block(
                512, num_classes, dropout=dropout_aux, key=keys[12]
            )
            self.aux2 = inception_aux_block(
                528, num_classes, dropout=dropout_aux, key=keys[13]
            )

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=dropout)
        self.fc = nn.Linear(1024, num_classes, key=keys[14])

    def __call__(
        self, x: Array, *, key: "jax.random.PRNGKey"
    ) -> Union[Array, Optional[Array], Optional[Array]]:
        """**Arguments:**

        - `x`: The input. Should be a JAX array with `3` channels
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        if key is None:
            raise RuntimeError("The model requires a PRNGKey.")
        keys = jrandom.split(key, 14)
        # N x 3 x 224 x 224
        x = self.conv1(x, key=keys[0])
        # N x 64 x 112 x 112
        x = self.maxpool1(x, key=keys[1])
        # N x 64 x 56 x 56
        x = self.conv2(x, key=keys[2])
        # N x 64 x 56 x 56
        x = self.conv3(x, key=keys[3])
        # N x 192 x 56 x 56
        x = self.maxpool2(x)

        # N x 192 x 28 x 28
        x = self.inception3a(x, key=keys[4])
        # N x 256 x 28 x 28
        x = self.inception3b(x, key=keys[5])
        # N x 480 x 28 x 28
        x = self.maxpool3(x)
        # N x 480 x 14 x 14
        x = self.inception4a(x, key=keys[6])
        # N x 512 x 14 x 14
        if self.aux_logits:
            aux1 = self.aux1(x, key=keys[7])

        x = self.inception4b(x, key=keys[8])
        # N x 512 x 14 x 14
        x = self.inception4c(x, key=keys[9])
        # N x 512 x 14 x 14
        x = self.inception4d(x, key=keys[10])
        # N x 528 x 14 x 14
        if self.aux_logits:
            aux2 = self.aux2(x, key=keys[11])  # Key here, a bad thing?

        x = self.inception4e(x, key=keys[12])
        # N x 832 x 14 x 14
        x = self.maxpool4(x)
        # N x 832 x 7 x 7
        x = self.inception5a(x, key=keys[13])
        # N x 832 x 7 x 7
        x = self.inception5b(x, key=keys[14])
        # N x 1024 x 7 x 7

        x = self.avgpool(x)
        # N x 1024 x 1 x 1
        x = jnp.ravel(x)
        # N x 1024
        x = self.dropout(x, key=keys[15])
        x = self.fc(x)
        # N x 1000 (num_classes)
        if self.aux_logits:
            return x, aux2, aux1
        else:
            return x


class _Inception(eqx.Module):
    branch1: eqx.Module
    branch2: nn.Sequential
    branch3: nn.Sequential
    branch4: nn.Sequential

    def __init__(
        self,
        in_channels: int,
        ch1x1: int,
        ch3x3red: int,
        ch3x3: int,
        ch5x5red: int,
        ch5x5: int,
        pool_proj: int,
        conv_block: Optional[Callable[..., eqx.Module]] = None,
        *,
        key: jax.random.PRNGKey = None
    ) -> None:
        super().__init__()
        if conv_block is None:
            conv_block = BasicConv2d
        keys = jrandom.split(key, 5)
        self.branch1 = conv_block(in_channels, ch1x1, kernel_size=1, key=keys[0])
        self.branch2 = nn.Sequential(
            [
                conv_block(in_channels, ch3x3red, kernel_size=1, key=keys[1]),
                conv_block(ch3x3red, ch3x3, kernel_size=3, padding=1, key=keys[2]),
            ]
        )

        self.branch3 = nn.Sequential(
            [
                conv_block(in_channels, ch5x5red, kernel_size=1, key=keys[3]),
                # Here, kernel_size=3 instead of kernel_size=5 is a known bug.
                # Please see https://github.com/pytorch/vision/issues/906 for details.
                conv_block(ch5x5red, ch5x5, kernel_size=3, padding=1, key=keys[4]),
            ]
        )

        self.branch4 = nn.Sequential(
            [
                nn.MaxPool2d(kernel_size=3, stride=1, padding=1, use_ceil=True),
                conv_block(in_channels, pool_proj, kernel_size=1, key=keys[5]),
            ]
        )

    def __call__(self, x: Array, *, key: jax.random.PRNGKey = None) -> Array:
        keys = jrandom.split(key, 4)
        branch1 = self.branch1(x, key=keys[0])
        branch2 = self.branch2(x, key=keys[1])
        branch3 = self.branch3(x, key=keys[2])
        branch4 = self.branch4(x, key=keys[3])

        outputs = jnp.concatenate([branch1, branch2, branch3, branch4], axis=0)
        return outputs


class InceptionAux(eqx.Module):
    conv: eqx.Module
    fc1: nn.Linear
    fc2: nn.Linear
    dropout: nn.Dropout
    avgpool: nn.AdaptiveAvgPool2d

    def __init__(
        self,
        in_channels: int,
        num_classes: int,
        conv_block: Optional[Callable[..., eqx.Module]] = None,
        dropout: float = 0.7,
        *,
        key: jax.random.PRNGKey = None
    ) -> None:
        super().__init__()
        if conv_block is None:
            conv_block = BasicConv2d
        keys = jrandom.split(key, 3)
        self.conv = conv_block(in_channels, 128, kernel_size=1, key=keys[0])
        self.fc1 = nn.Linear(2048, 1024, key=keys[1])
        self.fc2 = nn.Linear(1024, num_classes, key=keys[2])
        self.dropout = nn.Dropout(p=dropout)
        self.avgpool = nn.AdaptiveAvgPool2d(
            (4, 4),
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey" = None) -> Array:
        keys = jrandom.split(key, 2)
        # aux1: N x 512 x 14 x 14, aux2: N x 528 x 14 x 14
        x = self.avgpool(x)
        # aux1: N x 512 x 4 x 4, aux2: N x 528 x 4 x 4
        x = self.conv(x, key=keys[0])
        # N x 128 x 4 x 4
        x = jnp.ravel(x)
        # N x 2048
        x = jnn.relu(self.fc1(x))
        # N x 1024
        x = self.dropout(x, key=keys[1])
        # N x 1024
        x = self.fc2(x)
        # N x 1000 (num_classes)

        return x


class BasicConv2d(eqx.Module):
    conv: nn.Conv2d
    bn: eqx.experimental.BatchNorm

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        key: jax.random.PRNGKey = None,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels, out_channels, use_bias=False, key=key, **kwargs
        )
        self.bn = eqx.experimental.BatchNorm(out_channels, axis_name="batch", eps=0.001)

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        x = self.conv(x)
        x = self.bn(x, key=key)
        return jnn.relu(x)


def googlenet(pretrained=False, **kwargs: Any) -> GoogLeNet:
    r"""GoogLeNet (Inception v1) model architecture from
    [Going Deeper with Convolutions](http://arxiv.org/abs/1409.4842).
    The required minimum input size of the model is 15x15.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    if pretrained:
        use_aux = kwargs.get("aux_logits", False)
        model = GoogLeNet(aux_logits=True, **kwargs)
        model = load_torch_weights(model, url=MODEL_URLS["googlenet"])
        if not use_aux:
            model = eqx.tree_at(lambda m: m.aux_logits, model, replace=(False))
        else:
            warnings.warn(
                "Loaded pretrained weights for GoogLeNet. But, aux-branch weights are un-trained."
            )
    else:
        model = GoogLeNet(**kwargs)
    return model



from functools import partial
from typing import Callable, Optional

import equinox.experimental as eqxex
import equinox.nn as nn
import jax
import jax.nn as jnn


class ConvNormActivation(nn.Sequential):
    """A simple port of `torchvision.ops.misc.ConvNormActivation`.

    Packs `convolution` -> `normalisation` -> `activation` into one easy to use module.
    """

    out_channels: int

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: Optional[int] = None,
        groups: int = 1,
        norm_layer: Optional[Callable] = eqxex.BatchNorm,
        activation_layer: Optional[Callable] = jnn.relu,
        dilation: int = 1,
        use_bias: Optional[bool] = None,
        *,
        key: "jax.random.PRNGKey" = None
    ) -> None:
        """

        - `in_channels`: Number of channels in the input image
        - `out_channels`: Number of channels produced by the Convolution-Normalzation-Activation block
        - `kernel_size`: Size of the convolution kernel. Defaults to `3`
        - `stride`: Stride of the convolution. Defaults to `1`
        - `padding`: Padding added to all four sides of the input. Defaults to `None`,
            in which case it will calculated as ``padding = (kernel_size - 1) // 2 * dilation``
        - `groups`: Number of blocked connections from input channels to output channels. Defaults to `1`
        - `norm_layer`: Norm layer that will be stacked on top of the convolution layer. If ``None``
            this layer wont be used. Defaults to ``eqx.experimental.BatchNorm``
        - `activation_layer`: Activation function which will be stacked on top of the normalization layer
            (if not None), otherwise on top of the conv layer
            If ``None`` this layer wont be used. Defaults to ``jax.nn.relu``
        - `dilation`: Spacing between kernel elements. Defaults to `1`
        - `bias`: If `True`, bias is used in the convolution layer. By default, biases are included #
            if ``norm_layer is None``
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        if key is None:
            key = jax.random.PRNGKey(0)

        if padding is None:
            padding = (kernel_size - 1) // 2 * dilation
        if use_bias is None:
            use_bias = norm_layer is None
        layers = [
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                dilation=dilation,
                groups=groups,
                use_bias=use_bias,
                key=key,
            )
        ]
        if norm_layer is not None:
            is_bn = (
                norm_layer.func == eqxex.BatchNorm
                if isinstance(norm_layer, partial)
                else norm_layer == eqxex.BatchNorm
            )
            if is_bn:
                layers.append(norm_layer(out_channels, axis_name="batch"))
            else:
                layers.append(norm_layer(out_channels))
        if activation_layer is not None:
            layers.append(nn.Lambda(activation_layer))
        super().__init__(layers)
        self.out_channels = out_channels



from typing import Any, Optional, Sequence, Tuple, Union

import equinox as eqx
import equinox.experimental as eqxex
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...utils import load_torch_weights, MODEL_URLS


class _DenseLayer(eqx.Module):
    norm1: eqxex.BatchNorm
    relu: nn.Lambda
    conv1: nn.Conv2d
    norm2: eqxex.BatchNorm
    conv2: nn.Conv2d
    dropout: nn.Dropout

    def __init__(
        self,
        num_input_features: int,
        growth_rate: int,
        bn_size: int,
        drop_rate: float,
        key: "jax.random.PRNGKey",
    ) -> None:
        super().__init__()
        keys = jrandom.split(key, 2)
        self.norm1 = eqxex.BatchNorm(num_input_features, axis_name="batch")
        self.relu = nn.Lambda(jnn.relu)
        self.conv1 = nn.Conv2d(
            num_input_features,
            bn_size * growth_rate,
            kernel_size=1,
            stride=1,
            use_bias=False,
            key=keys[0],
        )
        self.norm2 = eqxex.BatchNorm(bn_size * growth_rate, axis_name="batch")
        self.conv2 = nn.Conv2d(
            bn_size * growth_rate,
            growth_rate,
            kernel_size=3,
            stride=1,
            padding=1,
            use_bias=False,
            key=keys[1],
        )
        self.dropout = nn.Dropout(p=float(drop_rate))

    def __call__(
        self, x: Union[Array, Sequence[Array]], *, key: "jax.random.PRNGKey"
    ) -> Array:
        if isinstance(x, Array):
            prev_features = [x]
        else:
            prev_features = x

        concated_features = jnp.concatenate(prev_features, axis=0)
        bottleneck_output = self.conv1(self.relu(self.norm1(concated_features)))
        new_features = self.conv2(self.relu(self.norm2(bottleneck_output)))
        new_features = self.dropout(new_features, key=key)
        return new_features


class _DenseBlock(eqx.Module):
    layers: Sequence[eqx.Module]
    num_layers: int

    def __init__(
        self,
        num_layers: int,
        num_input_features: int,
        bn_size: int,
        growth_rate: int,
        drop_rate: float,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        super().__init__()
        self.layers = []
        self.num_layers = num_layers
        keys = jrandom.split(key, num_layers)
        for i in range(num_layers):
            layer = _DenseLayer(
                num_input_features + i * growth_rate,
                growth_rate=growth_rate,
                bn_size=bn_size,
                drop_rate=drop_rate,
                key=keys[i],
            )
            self.layers.append(layer)

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        features = [x]
        keys = jrandom.split(key, self.num_layers)
        for i in range(self.num_layers):
            new_features = self.layers[i](features, key=keys[i])
            features.append(new_features)
        return jnp.concatenate(features, 0)


class _Transition(eqx.Module):
    layers: nn.Sequential

    def __init__(
        self,
        num_input_features: int,
        num_output_features: int,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        super().__init__()
        self.layers = nn.Sequential(
            [
                eqxex.BatchNorm(num_input_features, axis_name="batch"),
                nn.Lambda(jnn.relu),
                nn.Conv2d(
                    num_input_features,
                    num_output_features,
                    kernel_size=1,
                    stride=1,
                    use_bias=False,
                    key=key,
                ),
                nn.AvgPool2d(kernel_size=2, stride=2),
            ]
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        return self.layers(x, key=key)


class DenseNet(eqx.Module):
    """A simple port of `torchvision.models.densenet`"""

    features: nn.Sequential
    classifier: nn.Linear

    def __init__(
        self,
        growth_rate: int = 32,
        block_config: Tuple[int, int, int, int] = (6, 12, 24, 16),
        num_init_features: int = 64,
        bn_size: int = 4,
        drop_rate: float = 0,
        num_classes: int = 1000,
        *,
        key: Optional["jax.random.PRNGKey"] = None,
    ) -> None:
        """
        **Arguments:**

        -   `growth_rate`: Number of filters to add in each layer (`k` in paper)
        -   `block_config`: Number of layers in each pooling block
        -   `num_init_features` - The number of filters to learn in the first convolution layer
        -   `bn_size`: Multiplicative factor for number of bottle neck layers
              (i.e. bn_size * k features in the bottleneck layer)
        -   `drop_rate`: Dropout rate after each dense layer
        - `num_classes`: Number of classes in the classification task.
                            Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        """
        super().__init__()
        if key is None:
            key = jrandom.PRNGKey(0)
        # First convolution
        keys = jrandom.split(key, 2 * len(block_config) + 2)
        self.features = nn.Sequential(
            [
                nn.Conv2d(
                    3,
                    num_init_features,
                    kernel_size=7,
                    stride=2,
                    padding=3,
                    use_bias=False,
                    key=keys[0],
                ),
                eqxex.BatchNorm(num_init_features, axis_name="batch"),
                nn.Lambda(jnn.relu),
                nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
            ]
        )

        # Each denseblock
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            keys = jrandom.split(keys[i * 2 + 1], 3)
            block = _DenseBlock(
                num_layers=num_layers,
                num_input_features=num_features,
                bn_size=bn_size,
                growth_rate=growth_rate,
                drop_rate=drop_rate,
                key=keys[0],
            )
            self.features.layers.append(block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = _Transition(
                    num_input_features=num_features,
                    num_output_features=num_features // 2,
                    key=keys[i * 2 + 2],
                )
                self.features.layers.append(trans)
                num_features = num_features // 2

        # Final batch norm, relu and pooling
        self.features.layers.extend(
            [
                eqxex.BatchNorm(num_features, axis_name="batch"),
                nn.Lambda(jnn.relu),
                nn.AdaptiveAvgPool2d((1, 1)),
            ]
        )
        # Linear layer
        self.classifier = nn.Linear(num_features, num_classes, key=keys[-1])

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array with `3` channels
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        out = self.features(x, key=key)
        out = jnp.ravel(out)
        out = self.classifier(out)
        return out


def _densenet(
    growth_rate: int,
    block_config: Tuple[int, int, int, int],
    num_init_features: int,
    **kwargs: Any,
) -> DenseNet:
    model = DenseNet(growth_rate, block_config, num_init_features, **kwargs)
    return model


def densenet121(pretrained=False, **kwargs: Any) -> DenseNet:
    r"""Densenet-121 model from
    [Densely Connected Convolutional Networks](https://arxiv.org/pdf/1608.06993.pdf).
    The required minimum input size of the model is 29x29.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.


    """
    model = _densenet(32, (6, 12, 24, 16), 64, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["densenet121"])
    return model


def densenet161(pretrained=False, **kwargs: Any) -> DenseNet:
    r"""Densenet-161 model from
    [Densely Connected Convolutional Networks](https://arxiv.org/pdf/1608.06993.pdf).
    The required minimum input size of the model is 29x29.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.


    """
    model = _densenet(48, (6, 12, 36, 24), 96, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["densenet161"])
    return model


def densenet169(pretrained=False, **kwargs: Any) -> DenseNet:
    r"""Densenet-169 model from
    [Densely Connected Convolutional Networks](https://arxiv.org/pdf/1608.06993.pdf).
    The required minimum input size of the model is 29x29.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _densenet(32, (6, 12, 32, 32), 64, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["densenet169"])
    return model


def densenet201(pretrained=False, **kwargs: Any) -> DenseNet:
    r"""Densenet-201 model from
    [Densely Connected Convolutional Networks](https://arxiv.org/pdf/1608.06993.pdf).
    The required minimum input size of the model is 29x29.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _densenet(32, (6, 12, 48, 32), 64, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["densenet201"])
    return model



from typing import Any, Optional

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...utils import load_torch_weights, MODEL_URLS


class _Fire(eqx.Module):
    inplanes: int
    squeeze: nn.Conv2d
    squeeze_activation: nn.Lambda
    expand1x1: nn.Conv2d
    expand1x1_activation: nn.Lambda
    expand3x3: nn.Conv2d
    expand3x3_activation: nn.Lambda

    def __init__(
        self,
        inplanes: int,
        squeeze_planes: int,
        expand1x1_planes: int,
        expand3x3_planes: int,
        key=None,
    ) -> None:
        super().__init__()
        keys = jrandom.split(key, 3)
        self.inplanes = inplanes
        self.squeeze = nn.Conv2d(inplanes, squeeze_planes, kernel_size=1, key=keys[0])
        self.squeeze_activation = nn.Lambda(jnn.relu)
        self.expand1x1 = nn.Conv2d(
            squeeze_planes, expand1x1_planes, kernel_size=1, key=keys[1]
        )
        self.expand1x1_activation = nn.Lambda(jnn.relu)
        self.expand3x3 = nn.Conv2d(
            squeeze_planes, expand3x3_planes, kernel_size=3, padding=1, key=keys[2]
        )
        self.expand3x3_activation = nn.Lambda(jnn.relu)

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        x = self.squeeze_activation(self.squeeze(x))
        return jnp.concatenate(
            [
                self.expand1x1_activation(self.expand1x1(x)),
                self.expand3x3_activation(self.expand3x3(x)),
            ],
            axis=0,
        )


class SqueezeNet(eqx.Module):
    """A simple port of `torchvision.models.squeezenet`"""

    features: nn.Sequential
    classifier: nn.Sequential

    def __init__(
        self,
        version: str = "1_0",
        num_classes: int = 1000,
        dropout: float = 0.5,
        *,
        key: Optional["jax.random.PRNGKey"] = None
    ) -> None:
        """**Arguments:**

        - `version`: Specifies the version of the network. Defaults to `1_0`
        - `num_classes`: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        - `dropout`: The probability parameter for `equinox.nn.Dropout`
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        super().__init__()
        if key is None:
            key = jrandom.PRNGKey(0)
        keys = jrandom.split(key, 10)
        if version == "1_0":
            self.features = nn.Sequential(
                [
                    nn.Conv2d(3, 96, kernel_size=7, stride=2, key=keys[0]),
                    nn.Lambda(jnn.relu),
                    nn.MaxPool2d(kernel_size=3, stride=2, use_ceil=True),
                    _Fire(96, 16, 64, 64, key=keys[1]),
                    _Fire(128, 16, 64, 64, key=keys[2]),
                    _Fire(128, 32, 128, 128, key=keys[3]),
                    nn.MaxPool2d(kernel_size=3, stride=2, use_ceil=True),
                    _Fire(256, 32, 128, 128, key=keys[4]),
                    _Fire(256, 48, 192, 192, key=keys[5]),
                    _Fire(384, 48, 192, 192, key=keys[6]),
                    _Fire(384, 64, 256, 256, key=keys[7]),
                    nn.MaxPool2d(kernel_size=3, stride=2, use_ceil=True),
                    _Fire(512, 64, 256, 256, key=keys[8]),
                ]
            )
        elif version == "1_1":
            self.features = nn.Sequential(
                [
                    nn.Conv2d(3, 64, kernel_size=3, stride=2, key=keys[0]),
                    nn.Lambda(jnn.relu),
                    nn.MaxPool2d(kernel_size=3, stride=2, use_ceil=True),
                    _Fire(64, 16, 64, 64, key=keys[1]),
                    _Fire(128, 16, 64, 64, key=keys[2]),
                    nn.MaxPool2d(kernel_size=3, stride=2, use_ceil=True),
                    _Fire(128, 32, 128, 128, key=keys[3]),
                    _Fire(256, 32, 128, 128, key=keys[4]),
                    nn.MaxPool2d(kernel_size=3, stride=2, use_ceil=True),
                    _Fire(256, 48, 192, 192, key=keys[5]),
                    _Fire(384, 48, 192, 192, key=keys[6]),
                    _Fire(384, 64, 256, 256, key=keys[7]),
                    _Fire(512, 64, 256, 256, key=keys[8]),
                ]
            )

        # Final convolution is initialized differently from the rest
        final_conv = nn.Conv2d(512, num_classes, kernel_size=1, key=keys[9])
        self.classifier = nn.Sequential(
            [
                nn.Dropout(p=dropout),
                final_conv,
                nn.Lambda(jnn.relu),
                nn.AdaptiveAvgPool2d((1, 1)),
            ]
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array with `3` channels
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        x = self.features(x)
        x = self.classifier(x, key=key)
        return jnp.ravel(x)


def _squeezenet(version: str, pretrained: bool, **kwargs: Any) -> SqueezeNet:
    model = SqueezeNet(version, **kwargs)
    if pretrained:
        arch = "squeezenet" + version
        model = load_torch_weights(model, url=MODEL_URLS[arch])
    return model


def squeezenet1_0(pretrained: bool = False, **kwargs: Any) -> SqueezeNet:
    r"""SqueezeNet model architecture from the [SqueezeNet: AlexNet-level
    accuracy with 50x fewer parameters and <0.5MB model size](https://arxiv.org/abs/1602.07360) paper.
    The required minimum input size of the model is 21x21.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    return _squeezenet("1_0", pretrained, **kwargs)


def squeezenet1_1(pretrained: bool = False, **kwargs: Any) -> SqueezeNet:
    r"""SqueezeNet 1.1 model from the (official SqueezeNet repo)
    [https://github.com/DeepScale/SqueezeNet/tree/master/SqueezeNet_v1.1].
    SqueezeNet 1.1 has 2.4x less computation and slightly fewer parameters
    than SqueezeNet 1.0, without sacrificing accuracy.
    The required minimum input size of the model is 17x17.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    return _squeezenet("1_1", pretrained, **kwargs)



from .classification.alexnet import AlexNet, alexnet
from .classification.convnext import (
    ConvNeXt,
    convnext_base,
    convnext_large,
    convnext_small,
    convnext_tiny,
)
from .classification.densenet import (
    DenseNet,
    densenet121,
    densenet161,
    densenet169,
    densenet201,
)
from .classification.efficientnet import (
    EfficientNet,
    efficientnet_b0,
    efficientnet_b1,
    efficientnet_b2,
    efficientnet_b3,
    efficientnet_b4,
    efficientnet_b5,
    efficientnet_b6,
    efficientnet_b7,
    efficientnet_v2_l,
    efficientnet_v2_m,
    efficientnet_v2_s,
)
from .classification.googlenet import GoogLeNet, googlenet
from .classification.mobilenetv2 import mobilenet_v2, MobileNetV2
from .classification.mobilenetv3 import (
    mobilenet_v3_large,
    mobilenet_v3_small,
    MobileNetV3,
)
from .classification.regnet import (
    RegNet,
    regnet_x_1_6gf,
    regnet_x_3_2gf,
    regnet_x_8gf,
    regnet_x_16gf,
    regnet_x_32gf,
    regnet_x_400mf,
    regnet_x_800mf,
    regnet_y_1_6gf,
    regnet_y_3_2gf,
    regnet_y_8gf,
    regnet_y_16gf,
    regnet_y_32gf,
    regnet_y_128gf,
    regnet_y_400mf,
    regnet_y_800mf,
)
from .classification.resnet import (
    ResNet,
    resnet18,
    resnet34,
    resnet50,
    resnet101,
    resnet152,
    resnext50_32x4d,
    resnext101_32x8d,
    wide_resnet50_2,
    wide_resnet101_2,
)
from .classification.shufflenetv2 import (
    shufflenet_v2_x0_5,
    shufflenet_v2_x1_0,
    shufflenet_v2_x1_5,
    shufflenet_v2_x2_0,
    ShuffleNetV2,
)
from .classification.squeezenet import SqueezeNet, squeezenet1_0, squeezenet1_1
from .classification.swin import (
    swin_b,
    swin_s,
    swin_t,
    swin_v2_b,
    swin_v2_s,
    swin_v2_t,
    SwinTransformer,
)
from .classification.vgg import (
    VGG,
    vgg11,
    vgg11_bn,
    vgg13,
    vgg13_bn,
    vgg16,
    vgg16_bn,
    vgg19,
    vgg19_bn,
)
from .classification.vit import (
    _VitAttention,
    _VitBlock,
    VisionTransformer,
    vit_base,
    vit_small,
    vit_tiny,
)



from functools import partial
from typing import Any, Callable, List, Optional, Tuple

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jr
from equinox.custom_types import Array

from ...layers import ConvNormActivation, SqueezeExcitation
from ...utils import _make_divisible, load_torch_weights, MODEL_URLS


class SimpleStemIN(ConvNormActivation):
    """Simple stem for ImageNet: 3x3, BN, ReLU."""

    def __init__(
        self,
        width_in: int,
        width_out: int,
        norm_layer: Optional[Callable],
        activation_layer: Optional[Callable],
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        super().__init__(
            width_in,
            width_out,
            kernel_size=3,
            stride=2,
            norm_layer=norm_layer,
            activation_layer=activation_layer,
            key=key,
        )


class BottleneckTransform(nn.Sequential):
    """Bottleneck transformation: 1x1, 3x3 [+SE], 1x1."""

    def __init__(
        self,
        width_in: int,
        width_out: int,
        stride: int,
        norm_layer: Optional[Callable],
        activation_layer: Optional[Callable],
        group_width: int,
        bottleneck_multiplier: float,
        se_ratio: Optional[float],
        *,
        key: "jax.random.PRNGKey",
    ) -> None:
        layers = []
        keys = jr.split(key, 4)
        w_b = int(round(width_out * bottleneck_multiplier))
        g = w_b // group_width

        layers.append(
            ConvNormActivation(
                width_in,
                w_b,
                kernel_size=1,
                stride=1,
                norm_layer=norm_layer,
                activation_layer=activation_layer,
                key=keys[0],
            )
        )
        layers.append(
            ConvNormActivation(
                w_b,
                w_b,
                kernel_size=3,
                stride=stride,
                groups=g,
                norm_layer=norm_layer,
                activation_layer=activation_layer,
                key=keys[1],
            )
        )

        if se_ratio:
            # The SE reduction ratio is defined with respect to the
            # beginning of the block
            width_se_out = int(round(se_ratio * width_in))
            layers.append(
                SqueezeExcitation(
                    input_channels=w_b,
                    squeeze_channels=width_se_out,
                    activation=activation_layer,
                    key=keys[2],
                )
            )

        layers.append(
            ConvNormActivation(
                w_b,
                width_out,
                kernel_size=1,
                stride=1,
                norm_layer=norm_layer,
                activation_layer=None,
                key=keys[3],
            )
        )
        super().__init__(layers)


class ResBottleneckBlock(eqx.Module):
    """Residual bottleneck block: x + F(x), F = bottleneck transform."""

    proj: eqx.Module
    f: eqx.Module
    activation: Callable

    def __init__(
        self,
        width_in: int,
        width_out: int,
        stride: int,
        norm_layer: Optional[Callable],
        activation_layer: Optional[Callable],
        group_width: int = 1,
        bottleneck_multiplier: float = 1.0,
        se_ratio: Optional[float] = None,
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        super().__init__()

        # Use skip connection with projection if shape changes
        self.proj = nn.Identity()
        should_proj = (width_in != width_out) or (stride != 1)
        keys = jr.split(key, 2)
        if should_proj:
            self.proj = ConvNormActivation(
                width_in,
                width_out,
                kernel_size=1,
                stride=stride,
                norm_layer=norm_layer,
                activation_layer=None,
                key=keys[0],
            )
        self.f = BottleneckTransform(
            width_in,
            width_out,
            stride,
            norm_layer,
            activation_layer,
            group_width,
            bottleneck_multiplier,
            se_ratio,
            key=keys[1],
        )
        self.activation = activation_layer

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        keys = jr.split(key, 2)
        x = self.proj(x, key=keys[0]) + self.f(x, key=keys[1])
        return self.activation(x)


class AnyStage(nn.Sequential):
    """AnyNet stage (sequence of blocks w/ the same output shape)."""

    def __init__(
        self,
        width_in: int,
        width_out: int,
        stride: int,
        depth: int,
        block_constructor: eqx.Module,
        norm_layer: Callable,
        activation_layer: Callable,
        group_width: int,
        bottleneck_multiplier: float,
        se_ratio: Optional[float] = None,
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        keys = jr.split(key, depth)
        blocks = []
        for i in range(depth):
            blocks.append(
                block_constructor(
                    width_in if i == 0 else width_out,
                    width_out,
                    stride if i == 0 else 1,
                    norm_layer,
                    activation_layer,
                    group_width,
                    bottleneck_multiplier,
                    se_ratio,
                    key=keys[i],
                )
            )

        super().__init__(blocks)


class BlockParams:
    def __init__(
        self,
        depths: List[int],
        widths: List[int],
        group_widths: List[int],
        bottleneck_multipliers: List[float],
        strides: List[int],
        se_ratio: Optional[float] = None,
    ) -> None:
        self.depths = depths
        self.widths = widths
        self.group_widths = group_widths
        self.bottleneck_multipliers = bottleneck_multipliers
        self.strides = strides
        self.se_ratio = se_ratio

    @classmethod
    def from_init_params(
        cls,
        depth: int,
        w_0: int,
        w_a: float,
        w_m: float,
        group_width: int,
        bottleneck_multiplier: float = 1.0,
        se_ratio: Optional[float] = None,
    ) -> "BlockParams":
        """
        Programatically compute all the per-block settings,
        given the RegNet parameters.
        The first step is to compute the quantized linear block parameters,
        in log space. Key parameters are:
        - `w_a` is the width progression slope
        - `w_0` is the initial width
        - `w_m` is the width stepping in the log space
        In other terms
        `log(block_width) = log(w_0) + w_m * block_capacity`,
        with `bock_capacity` ramping up following the w_0 and w_a params.
        This block width is finally quantized to multiples of 8.
        The second step is to compute the parameters per stage,
        taking into account the skip connection and the final 1x1 convolutions.
        We use the fact that the output width is constant within a stage.
        """

        QUANT = 8
        STRIDE = 2

        if w_a < 0 or w_0 <= 0 or w_m <= 1 or w_0 % 8 != 0:
            raise ValueError("Invalid RegNet settings")
        # Compute the block widths. Each stage has one unique block width
        widths_cont = jnp.arange(depth) * w_a + w_0
        block_capacity = jnp.round(jnp.log(widths_cont / w_0) / jnp.log(w_m))
        block_widths = (
            (jnp.round(jnp.divide(w_0 * jnp.power(w_m, block_capacity), QUANT)) * QUANT)
            .astype(jnp.int32)
            .tolist()
        )
        num_stages = len(set(block_widths))

        # Convert to per stage parameters
        split_helper = zip(
            block_widths + [0],
            [0] + block_widths,
            block_widths + [0],
            [0] + block_widths,
        )
        splits = [w != wp or r != rp for w, wp, r, rp in split_helper]

        stage_widths = [w for w, t in zip(block_widths, splits[:-1]) if t]
        stage_depths = (
            jnp.diff(jnp.asarray([d for d, t in enumerate(splits) if t]))
            .astype(jnp.int32)
            .tolist()
        )

        strides = [STRIDE] * num_stages
        bottleneck_multipliers = [bottleneck_multiplier] * num_stages
        group_widths = [group_width] * num_stages

        # Adjust the compatibility of stage widths and group widths
        stage_widths, group_widths = cls._adjust_widths_groups_compatibilty(
            stage_widths, bottleneck_multipliers, group_widths
        )

        return cls(
            depths=stage_depths,
            widths=stage_widths,
            group_widths=group_widths,
            bottleneck_multipliers=bottleneck_multipliers,
            strides=strides,
            se_ratio=se_ratio,
        )

    def _get_expanded_params(self):
        return zip(
            self.widths,
            self.strides,
            self.depths,
            self.group_widths,
            self.bottleneck_multipliers,
        )

    @staticmethod
    def _adjust_widths_groups_compatibilty(
        stage_widths: List[int], bottleneck_ratios: List[float], group_widths: List[int]
    ) -> Tuple[List[int], List[int]]:
        """
        Adjusts the compatibility of widths and groups,
        depending on the bottleneck ratio.
        """
        # Compute all widths for the current settings
        widths = [int(w * b) for w, b in zip(stage_widths, bottleneck_ratios)]
        group_widths_min = [min(g, w_bot) for g, w_bot in zip(group_widths, widths)]

        # Compute the adjusted widths so that stage and group widths fit
        ws_bot = [
            _make_divisible(w_bot, g) for w_bot, g in zip(widths, group_widths_min)
        ]
        stage_widths = [int(w_bot / b) for w_bot, b in zip(ws_bot, bottleneck_ratios)]
        return stage_widths, group_widths_min


class RegNet(eqx.Module):
    """A simple port of `torchvision.models.regnet`"""

    stem: eqx.Module
    trunk_output: nn.Sequential
    avgpool: nn.AdaptiveAvgPool2d
    fc: eqx.Module

    def __init__(
        self,
        block_params: BlockParams,
        num_classes: int = 1000,
        stem_width: int = 32,
        stem_type: Optional[eqx.Module] = None,
        block_type: Optional[eqx.Module] = None,
        norm_layer: Optional[eqx.Module] = None,
        activation: Optional[Callable] = None,
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        """**Arguments:**

        - `block_params`: Configuration for the building blocks of the network
        - `num_classes`: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        - `stem_width`: Width of stems in the model
        - `stem_type`: Block type for the stems
        - `block_type`: Type of block to be used in building the model
        - `norm_layer`: Normalisation to be applied on the inputs. Defaults to `BatchNorm`
        - `activation`: Activation to be applied to the intermediate outputs. Defaults to `jax.nn.relu`
        - `key`:         - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
        """

        super().__init__()

        if stem_type is None:
            stem_type = SimpleStemIN
        if norm_layer is None:
            norm_layer = eqx.experimental.BatchNorm
        if block_type is None:
            block_type = ResBottleneckBlock
        if activation is None:
            activation = jnn.relu

        if key is None:
            key = jr.PRNGKey(0)
        keys = jr.split(key, 2)
        # Ad hoc stem
        self.stem = stem_type(
            3, stem_width, norm_layer, activation, key=keys[0]  # width_in
        )

        current_width = stem_width

        blocks = []
        for i, (
            width_out,
            stride,
            depth,
            group_width,
            bottleneck_multiplier,
        ) in enumerate(block_params._get_expanded_params()):
            keys = jr.split(keys[1], 2)
            blocks.append(
                AnyStage(
                    current_width,
                    width_out,
                    stride,
                    depth,
                    block_type,
                    norm_layer,
                    activation,
                    group_width,
                    bottleneck_multiplier,
                    block_params.se_ratio,
                    key=keys[0],
                ),
            )

            current_width = width_out

        self.trunk_output = nn.Sequential(blocks)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(
            in_features=current_width, out_features=num_classes, key=keys[1]
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array with `3` channels
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        keys = jr.split(key, 2)
        x = self.stem(x, key=keys[0])
        x = self.trunk_output(x, key=keys[1])
        x = self.avgpool(x)
        x = jnp.ravel(x)
        x = self.fc(x)
        return x


def _regnet(
    arch: str,
    block_params: BlockParams,
    pretrained: bool,
    **kwargs: Any,
) -> RegNet:

    norm_layer = kwargs.pop(
        "norm_layer", partial(eqx.experimental.BatchNorm, eps=1e-05, momentum=0.1)
    )
    model = RegNet(block_params, norm_layer=norm_layer, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS[arch])
    return model


def regnet_y_400mf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """Constructs a RegNetY_400MF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=16, w_0=48, w_a=27.89, w_m=2.09, group_width=8, se_ratio=0.25
    )
    return _regnet(
        arch="regnet_y_400mf", block_params=params, pretrained=pretrained, **kwargs
    )


def regnet_y_800mf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetY_800MF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=14, w_0=56, w_a=38.84, w_m=2.4, group_width=16, se_ratio=0.25
    )
    return _regnet("regnet_y_800mf", params, pretrained, **kwargs)


def regnet_y_1_6gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetY_1.6GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=27, w_0=48, w_a=20.71, w_m=2.65, group_width=24, se_ratio=0.25
    )
    return _regnet("regnet_y_1_6gf", params, pretrained, **kwargs)


def regnet_y_3_2gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetY_3.2GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=21, w_0=80, w_a=42.63, w_m=2.66, group_width=24, se_ratio=0.25
    )
    return _regnet("regnet_y_3_2gf", params, pretrained, **kwargs)


def regnet_y_8gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetY_8GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=17, w_0=192, w_a=76.82, w_m=2.19, group_width=56, se_ratio=0.25
    )
    return _regnet("regnet_y_8gf", params, pretrained, **kwargs)


def regnet_y_16gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetY_16GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=18, w_0=200, w_a=106.23, w_m=2.48, group_width=112, se_ratio=0.25
    )
    return _regnet("regnet_y_16gf", params, pretrained, **kwargs)


def regnet_y_32gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetY_32GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=20, w_0=232, w_a=115.89, w_m=2.53, group_width=232, se_ratio=0.25
    )
    return _regnet("regnet_y_32gf", params, pretrained, **kwargs)


def regnet_y_128gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetY_128GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=27, w_0=456, w_a=160.83, w_m=2.52, group_width=264, se_ratio=0.25
    )
    return _regnet("regnet_y_128gf", params, pretrained, **kwargs)


def regnet_x_400mf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetX_400MF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=22, w_0=24, w_a=24.48, w_m=2.54, group_width=16
    )
    return _regnet("regnet_x_400mf", params, pretrained, **kwargs)


def regnet_x_800mf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetX_800MF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=16, w_0=56, w_a=35.73, w_m=2.28, group_width=16
    )
    return _regnet("regnet_x_800mf", params, pretrained, **kwargs)


def regnet_x_1_6gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetX_1.6GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=18, w_0=80, w_a=34.01, w_m=2.25, group_width=24
    )
    return _regnet("regnet_x_1_6gf", params, pretrained, **kwargs)


def regnet_x_3_2gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetX_3.2GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=25, w_0=88, w_a=26.31, w_m=2.25, group_width=48
    )
    return _regnet("regnet_x_3_2gf", params, pretrained, **kwargs)


def regnet_x_8gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetX_8GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=23, w_0=80, w_a=49.56, w_m=2.88, group_width=120
    )
    return _regnet("regnet_x_8gf", params, pretrained, **kwargs)


def regnet_x_16gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetX_16GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """

    params = BlockParams.from_init_params(
        depth=22, w_0=216, w_a=55.59, w_m=2.1, group_width=128
    )
    return _regnet("regnet_x_16gf", params, pretrained, **kwargs)


def regnet_x_32gf(pretrained: bool = False, **kwargs: Any) -> RegNet:
    """
    Constructs a RegNetX_32GF architecture from
    [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    params = BlockParams.from_init_params(
        depth=23, w_0=320, w_a=69.86, w_m=2.0, group_width=168
    )
    return _regnet("regnet_x_32gf", params, pretrained, **kwargs)



from typing import Any, Optional

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...utils import load_torch_weights, MODEL_URLS


class AlexNet(eqx.Module):
    """A simple port of `torchvision.models.alexnet`"""

    features: eqx.Module
    avgpool: eqx.Module
    classifier: eqx.Module

    def __init__(
        self,
        num_classes: int = 1000,
        dropout: float = 0.5,
        *,
        key: Optional["jax.random.PRNGKey"] = None
    ) -> None:
        """**Arguments:**

        - `num_classes`: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        - `dropout`: Parameter used for the `equinox.nn.Dropout` layers. Defaults to `0.5`
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)

        """
        super().__init__()
        if not key:
            key = jrandom.PRNGKey(0)
        keys = jrandom.split(key, 8)

        self.features = nn.Sequential(
            [
                nn.Conv2d(3, 64, kernel_size=11, stride=4, padding=2, key=keys[0]),
                nn.Lambda(jnn.relu),
                nn.MaxPool2d(kernel_size=3, stride=2),
                nn.Conv2d(64, 192, kernel_size=5, padding=2, key=keys[1]),
                nn.Lambda(jnn.relu),
                nn.MaxPool2d(kernel_size=3, stride=2),
                nn.Conv2d(192, 384, kernel_size=3, padding=1, key=keys[2]),
                nn.Lambda(jnn.relu),
                nn.Conv2d(384, 256, kernel_size=3, padding=1, key=keys[3]),
                nn.Lambda(jnn.relu),
                nn.Conv2d(256, 256, kernel_size=3, padding=1, key=keys[4]),
                nn.Lambda(jnn.relu),
                nn.MaxPool2d(kernel_size=3, stride=2),
            ]
        )
        self.avgpool = nn.AdaptiveAvgPool2d((6, 6))
        self.classifier = nn.Sequential(
            [
                nn.Dropout(p=dropout),
                nn.Linear(256 * 6 * 6, 4096, key=keys[5]),
                nn.Lambda(jnn.relu),
                nn.Dropout(p=dropout),
                nn.Linear(4096, 4096, key=keys[6]),
                nn.Lambda(jnn.relu),
                nn.Linear(4096, num_classes, key=keys[7]),
            ]
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array with `3` channels
        - `key`: Required parameter
        """
        if key is None:
            raise RuntimeError("The model requires a PRNGKey.")
        keys = jrandom.split(key, 2)
        x = self.features(x, key=keys[0])
        x = self.avgpool(x)
        x = jnp.ravel(x)
        x = self.classifier(x, key=keys[1])
        return x

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value


def alexnet(pretrained: bool = False, **kwargs: Any) -> AlexNet:
    r"""AlexNet model architecture from the
    [One weird trick...](https://arxiv.org/abs/1404.5997)` paper.
    The required minimum input size of the model is 63x63.
    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    model = AlexNet(**kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["alexnet"])
    return model



from typing import Any, Callable, List, Optional

import equinox as eqx
import equinox.experimental as eqxex
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...layers import ConvNormActivation
from ...utils import _make_divisible, load_torch_weights, MODEL_URLS


class _InvertedResidual(eqx.Module):
    stride: int
    use_res_connect: int
    conv: nn.Sequential
    out_channels: int

    def __init__(
        self,
        inp: int,
        oup: int,
        stride: int,
        expand_ratio: int,
        norm_layer: Optional[Callable[..., eqx.Module]] = None,
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        super().__init__()

        keys = jrandom.split(key, 3)

        self.stride = stride
        assert stride in [1, 2]

        if norm_layer is None:
            norm_layer = eqxex.BatchNorm

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup

        layers: List[eqx.Module] = []
        if expand_ratio != 1:
            # pw
            layers.append(
                ConvNormActivation(
                    inp,
                    hidden_dim,
                    kernel_size=1,
                    norm_layer=norm_layer,
                    activation_layer=jnn.relu,
                    key=keys[0],
                )
            )
        layers.extend(
            [
                # dw
                ConvNormActivation(
                    hidden_dim,
                    hidden_dim,
                    stride=stride,
                    groups=hidden_dim,
                    norm_layer=norm_layer,
                    activation_layer=jnn.relu,
                    key=keys[1],
                ),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, use_bias=False, key=keys[2]),
                norm_layer(oup, axis_name="batch"),
            ]
        )
        self.conv = nn.Sequential(layers)
        self.out_channels = oup

    def __call__(self, x, *, key: Optional["jax.random.PRNGKey"] = None) -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Forwarded to individual `eqx.Module` attributes
        """
        if self.use_res_connect:
            return x + self.conv(x, key=key)
        else:
            return self.conv(x, key=key)


class MobileNetV2(eqx.Module):
    """A simple port of `torchvision.models.mobilenetv2`"""

    features: nn.Sequential
    classifier: nn.Linear
    pool: nn.AdaptivePool

    def __init__(
        self,
        num_classes: int = 1000,
        width_mult: float = 1.0,
        inverted_residual_setting: Optional[List[List[int]]] = None,
        round_nearest: int = 8,
        block: Optional["eqx.Module"] = None,
        norm_layer: Optional["eqx.Module"] = None,
        dropout: float = 0.2,
        *,
        key: Optional["jax.random.PRNGKey"] = None,
    ) -> None:
        """**Arguments:**

        - `num_classes`: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        - `width_mult`: Adjusts number of channels in each layer by this amount
            inverted_residual_setting: Network structure
        - `inverted_residual_setting`: Network structure
        - `round_nearest`: Round the number of channels in each layer to be a multiple of this number
            Set to `1` to turn off rounding
        - `block`: Module specifying inverted residual building block for mobilenet
        - `norm_layer`: Module specifying the normalization layer to use
        - `dropout`: The dropout probability
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        super().__init__()

        if key is None:
            key = jrandom.PRNGKey(0)
        keys = jrandom.split(key, 2)

        if block is None:
            block = _InvertedResidual

        if norm_layer is None:
            norm_layer = eqxex.BatchNorm

        input_channel = 32
        last_channel = 1280

        if inverted_residual_setting is None:
            inverted_residual_setting = [
                # t, c, n, s
                [1, 16, 1, 1],
                [6, 24, 2, 2],
                [6, 32, 3, 2],
                [6, 64, 4, 2],
                [6, 96, 3, 1],
                [6, 160, 3, 2],
                [6, 320, 1, 1],
            ]

        # only check the first element, assuming user knows t,c,n,s are required
        if (
            len(inverted_residual_setting) == 0
            or len(inverted_residual_setting[0]) != 4
        ):
            raise ValueError(
                f"inverted_residual_setting should be non-empty or a 4-element list, got {inverted_residual_setting}"
            )

        # building first layer
        input_channel = _make_divisible(input_channel * width_mult, round_nearest)
        last_channel = _make_divisible(
            last_channel * max(1.0, width_mult), round_nearest
        )
        features: List[eqx.Module] = [
            ConvNormActivation(
                3,
                input_channel,
                stride=2,
                norm_layer=norm_layer,
                activation_layer=jnn.relu,
                key=keys[0],
            )
        ]
        # building inverted residual blocks
        for t, c, n, s in inverted_residual_setting:
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                keys = jrandom.split(keys[1], 2)
                stride = s if i == 0 else 1
                features.append(
                    block(
                        input_channel,
                        output_channel,
                        stride,
                        expand_ratio=t,
                        norm_layer=norm_layer,
                        key=keys[0],
                    )
                )
                input_channel = output_channel
        # building last several layers
        keys = jrandom.split(keys[1], 2)
        features.append(
            ConvNormActivation(
                input_channel,
                last_channel,
                kernel_size=1,
                norm_layer=norm_layer,
                activation_layer=jnn.relu,
                key=keys[0],
            )
        )
        # make it nn.Sequential
        self.features = nn.Sequential(features)

        # building classifier
        self.classifier = nn.Sequential(
            [
                nn.Dropout(p=dropout),
                nn.Linear(last_channel, num_classes, key=keys[1]),
            ]
        )

        self.pool = nn.AdaptiveAvgPool2d((1, 1))

    def __call__(self, x, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        keys = jrandom.split(key, 3)
        x = self.features(x, key=keys[0])
        x = self.pool(x, key=keys[1])
        x = jnp.ravel(x)
        x = self.classifier(x, key=keys[2])
        return x


def mobilenet_v2(pretrained=False, **kwargs: Any) -> MobileNetV2:
    """
    Constructs a MobileNetV2 architecture from
    [MobileNetV2: Inverted Residuals and Linear Bottlenecks](https://arxiv.org/abs/1801.04381).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = MobileNetV2(**kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["mobilenet_v2"])
    return model



from functools import partial
from typing import Any, Callable, List, Optional, Sequence

import equinox as eqx
import equinox.experimental as eqxex
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...layers import ConvNormActivation
from ...layers import SqueezeExcitation as SElayer
from ...utils import _make_divisible, load_torch_weights, MODEL_URLS


class _InvertedResidualConfig:
    # Stores information listed at Tables 1 and 2 of the MobileNetV3 paper
    def __init__(
        self,
        input_channels: int,
        kernel: int,
        expanded_channels: int,
        out_channels: int,
        use_se: bool,
        activation: str,
        stride: int,
        dilation: int,
        width_mult: float,
    ):
        self.input_channels = self.adjust_channels(input_channels, width_mult)
        self.kernel = kernel
        self.expanded_channels = self.adjust_channels(expanded_channels, width_mult)
        self.out_channels = self.adjust_channels(out_channels, width_mult)
        self.use_se = use_se
        self.use_hs = activation == "HS"
        self.stride = stride
        self.dilation = dilation

    @staticmethod
    def adjust_channels(channels: int, width_mult: float):
        return _make_divisible(channels * width_mult, 8)


class _InvertedResidual(eqx.Module):
    # Implemented as described at section 5 of MobileNetV3 paper
    use_res_connect: int
    block: nn.Sequential
    out_channels: int

    def __init__(
        self,
        cnf: _InvertedResidualConfig,
        norm_layer: Callable[..., eqx.Module],
        se_layer: Callable[..., eqx.Module] = partial(
            SElayer, scale_activation=jnn.hard_sigmoid
        ),
        *,
        key: "jax.random.PRNGKey" = None,
    ):
        super().__init__()
        keys = jrandom.split(key, 4)
        if not (1 <= cnf.stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = (
            cnf.stride == 1 and cnf.input_channels == cnf.out_channels
        )

        layers: List[eqx.Module] = []
        activation_layer = jnn.hard_swish if cnf.use_hs else jnn.relu

        # expand
        if cnf.expanded_channels != cnf.input_channels:
            layers.append(
                ConvNormActivation(
                    cnf.input_channels,
                    cnf.expanded_channels,
                    kernel_size=1,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                    key=keys[0],
                )
            )

        # depthwise
        stride = 1 if cnf.dilation > 1 else cnf.stride
        layers.append(
            ConvNormActivation(
                cnf.expanded_channels,
                cnf.expanded_channels,
                kernel_size=cnf.kernel,
                stride=stride,
                dilation=cnf.dilation,
                groups=cnf.expanded_channels,
                norm_layer=norm_layer,
                activation_layer=activation_layer,
                key=keys[1],
            )
        )
        if cnf.use_se:
            squeeze_channels = _make_divisible(cnf.expanded_channels // 4, 8)
            layers.append(
                se_layer(cnf.expanded_channels, squeeze_channels, key=keys[2])
            )

        # project
        layers.append(
            ConvNormActivation(
                cnf.expanded_channels,
                cnf.out_channels,
                kernel_size=1,
                norm_layer=norm_layer,
                activation_layer=None,
                key=keys[3],
            )
        )

        self.block = nn.Sequential(layers)
        self.out_channels = cnf.out_channels

    def __call__(self, x, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        result = self.block(x, key=key)
        if self.use_res_connect:
            result += x
        return result


class MobileNetV3(eqx.Module):
    """A simple port of `torchvision.models.mobilenetv3`"""

    features: nn.Sequential
    avgpool: nn.AdaptiveAvgPool2d
    classifier: nn.Sequential

    def __init__(
        self,
        inverted_residual_setting: List["_InvertedResidualConfig"],
        last_channel: int,
        num_classes: int = 1000,
        block: Optional["eqx.Module"] = None,
        norm_layer: Optional["eqx.Module"] = None,
        dropout: float = 0.2,
        *,
        key: Optional["jax.random.PRNGKey"] = None,
    ) -> None:
        """**Arguments:**

        - `inverted_residual_setting`: Network structure
        - `last_channel`: The number of channels on the penultimate layer
        - `num_classes`: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        - `block`: Module specifying inverted residual building block for mobilenet
        - `norm_layer`: Module specifying the normalization layer to use
        - `dropout`: The dropout probability
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        super().__init__()
        if key is None:
            key = jrandom.PRNGKey(0)

        keys = jrandom.split(key, 5)
        if not inverted_residual_setting:
            raise ValueError("The inverted_residual_setting should not be empty")
        elif not (
            isinstance(inverted_residual_setting, Sequence)
            and all(
                [
                    isinstance(s, _InvertedResidualConfig)
                    for s in inverted_residual_setting
                ]
            )
        ):
            raise TypeError(
                "The inverted_residual_setting should be List[InvertedResidualConfig]"
            )

        if block is None:
            block = _InvertedResidual

        if norm_layer is None:
            norm_layer = partial(eqxex.BatchNorm, eps=0.001, momentum=0.01)

        layers: List[eqx.Module] = []

        # building first layer
        firstconv_output_channels = inverted_residual_setting[0].input_channels
        layers.append(
            ConvNormActivation(
                3,
                firstconv_output_channels,
                kernel_size=3,
                stride=2,
                norm_layer=norm_layer,
                activation_layer=jnn.hard_swish,
                key=keys[0],
            )
        )

        # building inverted residual blocks
        for cnf in inverted_residual_setting:
            layers.append(block(cnf, norm_layer, key=keys[1]))

        # building last several layers
        lastconv_input_channels = inverted_residual_setting[-1].out_channels
        lastconv_output_channels = 6 * lastconv_input_channels
        layers.append(
            ConvNormActivation(
                lastconv_input_channels,
                lastconv_output_channels,
                kernel_size=1,
                norm_layer=norm_layer,
                activation_layer=jnn.hard_swish,
                key=keys[2],
            )
        )

        self.features = nn.Sequential(layers)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.classifier = nn.Sequential(
            [
                nn.Linear(lastconv_output_channels, last_channel, key=keys[3]),
                nn.Lambda(jnn.hard_swish),
                nn.Dropout(p=dropout),
                nn.Linear(last_channel, num_classes, key=keys[4]),
            ]
        )

    def __call__(self, x, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        keys = jrandom.split(key, 3)
        x = self.features(x, key=keys[0])
        x = self.avgpool(x, key=keys[1])
        x = jnp.ravel(x)
        x = self.classifier(x, key=keys[2])
        return x


def _mobilenet_v3_conf(
    arch: str,
    width_mult: float = 1.0,
    reduced_tail: bool = False,
    dilated: bool = False,
    **kwargs: Any,
):
    reduce_divider = 2 if reduced_tail else 1
    dilation = 2 if dilated else 1

    bneck_conf = partial(_InvertedResidualConfig, width_mult=width_mult)
    adjust_channels = partial(
        _InvertedResidualConfig.adjust_channels, width_mult=width_mult
    )

    if arch == "mobilenet_v3_large":
        inverted_residual_setting = [
            bneck_conf(16, 3, 16, 16, False, "RE", 1, 1),
            bneck_conf(16, 3, 64, 24, False, "RE", 2, 1),  # C1
            bneck_conf(24, 3, 72, 24, False, "RE", 1, 1),
            bneck_conf(24, 5, 72, 40, True, "RE", 2, 1),  # C2
            bneck_conf(40, 5, 120, 40, True, "RE", 1, 1),
            bneck_conf(40, 5, 120, 40, True, "RE", 1, 1),
            bneck_conf(40, 3, 240, 80, False, "HS", 2, 1),  # C3
            bneck_conf(80, 3, 200, 80, False, "HS", 1, 1),
            bneck_conf(80, 3, 184, 80, False, "HS", 1, 1),
            bneck_conf(80, 3, 184, 80, False, "HS", 1, 1),
            bneck_conf(80, 3, 480, 112, True, "HS", 1, 1),
            bneck_conf(112, 3, 672, 112, True, "HS", 1, 1),
            bneck_conf(
                112, 5, 672, 160 // reduce_divider, True, "HS", 2, dilation
            ),  # C4
            bneck_conf(
                160 // reduce_divider,
                5,
                960 // reduce_divider,
                160 // reduce_divider,
                True,
                "HS",
                1,
                dilation,
            ),
            bneck_conf(
                160 // reduce_divider,
                5,
                960 // reduce_divider,
                160 // reduce_divider,
                True,
                "HS",
                1,
                dilation,
            ),
        ]
        last_channel = adjust_channels(1280 // reduce_divider)  # C5
    elif arch == "mobilenet_v3_small":
        inverted_residual_setting = [
            bneck_conf(16, 3, 16, 16, True, "RE", 2, 1),  # C1
            bneck_conf(16, 3, 72, 24, False, "RE", 2, 1),  # C2
            bneck_conf(24, 3, 88, 24, False, "RE", 1, 1),
            bneck_conf(24, 5, 96, 40, True, "HS", 2, 1),  # C3
            bneck_conf(40, 5, 240, 40, True, "HS", 1, 1),
            bneck_conf(40, 5, 240, 40, True, "HS", 1, 1),
            bneck_conf(40, 5, 120, 48, True, "HS", 1, 1),
            bneck_conf(48, 5, 144, 48, True, "HS", 1, 1),
            bneck_conf(48, 5, 288, 96 // reduce_divider, True, "HS", 2, dilation),  # C4
            bneck_conf(
                96 // reduce_divider,
                5,
                576 // reduce_divider,
                96 // reduce_divider,
                True,
                "HS",
                1,
                dilation,
            ),
            bneck_conf(
                96 // reduce_divider,
                5,
                576 // reduce_divider,
                96 // reduce_divider,
                True,
                "HS",
                1,
                dilation,
            ),
        ]
        last_channel = adjust_channels(1024 // reduce_divider)  # C5
    else:
        raise ValueError(f"Unsupported model type {arch}")

    return inverted_residual_setting, last_channel


def _mobilenet_v3(
    arch: str,
    inverted_residual_setting: List[_InvertedResidualConfig],
    last_channel: int,
    **kwargs: Any,
):
    model = MobileNetV3(inverted_residual_setting, last_channel, **kwargs)
    return model


def mobilenet_v3_large(pretrained=False, **kwargs: Any) -> MobileNetV3:
    """
    Constructs a large MobileNetV3 architecture from
    [Searching for MobileNetV3](https://arxiv.org/abs/1905.02244).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    arch = "mobilenet_v3_large"
    inverted_residual_setting, last_channel = _mobilenet_v3_conf(arch, **kwargs)
    model = _mobilenet_v3(arch, inverted_residual_setting, last_channel, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["mobilenet_v3_large"])
    return model


def mobilenet_v3_small(pretrained=False, **kwargs: Any) -> MobileNetV3:
    """
    Constructs a small MobileNetV3 architecture from
    [Searching for MobileNetV3](https://arxiv.org/abs/1905.02244).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    arch = "mobilenet_v3_small"
    inverted_residual_setting, last_channel = _mobilenet_v3_conf(arch, **kwargs)
    model = _mobilenet_v3(arch, inverted_residual_setting, last_channel, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["mobilenet_v3_small"])
    return model



import logging
import os
import sys
import warnings
from pathlib import Path
from typing import NewType, Optional

import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu


try:
    import torch
except ImportError:
    warnings.warn("PyTorch is required for loading Torchvision pre-trained weights.")

_TEMP_DIR = "/tmp/.eqx"
_Url = NewType("_Url", str)
MODEL_URLS = {
    "alexnet": "https://download.pytorch.org/models/alexnet-owt-7be5be79.pth",
    "convnext_tiny": "https://download.pytorch.org/models/convnext_tiny-983f1562.pth",
    "convnext_small": "https://download.pytorch.org/models/convnext_small-0c510722.pth",
    "convnext_base": "https://download.pytorch.org/models/convnext_base-6075fbad.pth",
    "convnext_large": "https://download.pytorch.org/models/convnext_large-ea097f82.pth",
    "densenet121": "https://download.pytorch.org/models/densenet121-a639ec97.pth",
    "densenet169": "https://download.pytorch.org/models/densenet169-b2777c0a.pth",
    "densenet201": "https://download.pytorch.org/models/densenet201-c1103571.pth",
    "densenet161": "https://download.pytorch.org/models/densenet161-8d451a50.pth",
    "efficientnet_b0": "https://download.pytorch.org/models/efficientnet_b0_rwightman-3dd342df.pth",
    "efficientnet_b1": "https://download.pytorch.org/models/efficientnet_b1_rwightman-533bc792.pth",
    "efficientnet_b2": "https://download.pytorch.org/models/efficientnet_b2_rwightman-bcdf34b7.pth",
    "efficientnet_b3": "https://download.pytorch.org/models/efficientnet_b3_rwightman-cf984f9c.pth",
    "efficientnet_b4": "https://download.pytorch.org/models/efficientnet_b4_rwightman-7eb33cd5.pth",
    "efficientnet_b5": "https://download.pytorch.org/models/efficientnet_b5_lukemelas-b6417697.pth",
    "efficientnet_b6": "https://download.pytorch.org/models/efficientnet_b6_lukemelas-c76e70fd.pth",
    "efficientnet_b7": "https://download.pytorch.org/models/efficientnet_b7_lukemelas-dcc49843.pth",
    "efficientnet_v2_s": "https://download.pytorch.org/models/efficientnet_v2_s-dd5fe13b.pth",
    "efficientnet_v2_m": "https://download.pytorch.org/models/efficientnet_v2_m-dc08266a.pth",
    "efficientnet_v2_l": "https://download.pytorch.org/models/efficientnet_v2_l-59c71312.pth",
    "googlenet": "https://download.pytorch.org/models/googlenet-1378be20.pth",
    "mobilenet_v2": "https://download.pytorch.org/models/mobilenet_v2-b0353104.pth",
    "mobilenet_v3_large": "https://download.pytorch.org/models/mobilenet_v3_large-8738ca79.pth",
    "mobilenet_v3_small": "https://download.pytorch.org/models/mobilenet_v3_small-047dcff4.pth",
    "regnet_y_400mf": "https://download.pytorch.org/models/regnet_y_400mf-e6988f5f.pth",
    "regnet_y_800mf": "https://download.pytorch.org/models/regnet_y_800mf-58fc7688.pth",
    "regnet_y_1_6gf": "https://download.pytorch.org/models/regnet_y_1_6gf-0d7bc02a.pth",
    "regnet_y_3_2gf": "https://download.pytorch.org/models/regnet_y_3_2gf-9180c971.pth",
    "regnet_y_8gf": "https://download.pytorch.org/models/regnet_y_8gf-dc2b1b54.pth",
    "regnet_y_16gf": "https://download.pytorch.org/models/regnet_y_16gf-3e4a00f9.pth",
    "regnet_y_32gf": "https://download.pytorch.org/models/regnet_y_32gf-8db6d4b5.pth",
    "regnet_y_128gf": "https://download.pytorch.org/models/regnet_y_128gf_swag-c8ce3e52.pth",
    "regnet_x_400mf": "https://download.pytorch.org/models/regnet_x_400mf-62229a5f.pth",
    "regnet_x_800mf": "https://download.pytorch.org/models/regnet_x_800mf-94a99ebd.pth",
    "regnet_x_1_6gf": "https://download.pytorch.org/models/regnet_x_1_6gf-a12f2b72.pth",
    "regnet_x_3_2gf": "https://download.pytorch.org/models/regnet_x_3_2gf-7071aa85.pth",
    "regnet_x_8gf": "https://download.pytorch.org/models/regnet_x_8gf-2b70d774.pth",
    "regnet_x_16gf": "https://download.pytorch.org/models/regnet_x_16gf-ba3796d7.pth",
    "regnet_x_32gf": "https://download.pytorch.org/models/regnet_x_32gf-6eb8fdc6.pth",
    "resnet18": "https://download.pytorch.org/models/resnet18-5c106cde.pth",
    "resnet34": "https://download.pytorch.org/models/resnet34-333f7ec4.pth",
    "resnet50": "https://download.pytorch.org/models/resnet50-19c8e357.pth",
    "resnet101": "https://download.pytorch.org/models/resnet101-5d3b4d8f.pth",
    "resnet152": "https://download.pytorch.org/models/resnet152-b121ed2d.pth",
    "resnext50_32x4d": "https://download.pytorch.org/models/resnext50_32x4d-7cdf4587.pth",
    "resnext101_32x8d": "https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth",
    "shufflenetv2_x0.5": "https://download.pytorch.org/models/shufflenetv2_x0.5-f707e7126e.pth",
    "shufflenetv2_x1.0": "https://download.pytorch.org/models/shufflenetv2_x1-5666bf0f80.pth",
    "shufflenetv2_x1.5": None,
    "shufflenetv2_x2.0": None,
    "squeezenet1_0": "https://download.pytorch.org/models/squeezenet1_0-b66bff10.pth",
    "squeezenet1_1": "https://download.pytorch.org/models/squeezenet1_1-b8a52dc0.pth",
    "swin_t": "https://download.pytorch.org/models/swin_t-704ceda3.pth",
    "swin_s": "https://download.pytorch.org/models/swin_s-5e29d889.pth",
    "sim_b": "https://download.pytorch.org/models/swin_b-68c6b09e.pth",
    "swin_v2_t": "https://download.pytorch.org/models/swin_v2_t-b137f0e2.pth",
    "swin_v2_s": "https://download.pytorch.org/models/swin_v2_s-637d8ceb.pth",
    "sim_v2_b": "https://download.pytorch.org/models/swin_v2_b-781e5279.pth",
    "vit_small_patch16_224_dino": "https://dl.fbaipublicfiles.com/dino/"
    "dino_deitsmall16_pretrain/dino_deitsmall16_pretrain.pth",
    "vit_small_patch8_224_dino": "https://dl.fbaipublicfiles.com/dino/"
    "dino_deitsmall8_pretrain/dino_deitsmall8_pretrain.pth",
    "vit_base_patch16_224_dino": "https://dl.fbaipublicfiles.com/dino/"
    "dino_vitbase16_pretrain/dino_vitbase16_pretrain.pth",
    "vit_base_patch8_224_dino": "https://dl.fbaipublicfiles.com/dino/"
    "dino_vitbase8_pretrain/dino_vitbase8_pretrain.pth",
    "vgg11": "https://download.pytorch.org/models/vgg11-8a719046.pth",
    "vgg13": "https://download.pytorch.org/models/vgg13-19584684.pth",
    "vgg16": "https://download.pytorch.org/models/vgg16-397923af.pth",
    "vgg19": "https://download.pytorch.org/models/vgg19-dcbb9e9d.pth",
    "vgg11_bn": "https://download.pytorch.org/models/vgg11_bn-6002323d.pth",
    "vgg13_bn": "https://download.pytorch.org/models/vgg13_bn-abd245e5.pth",
    "vgg16_bn": "https://download.pytorch.org/models/vgg16_bn-6c64b313.pth",
    "vgg19_bn": "https://download.pytorch.org/models/vgg19_bn-c79401a0.pth",
    "wide_resnet50_2": "https://download.pytorch.org/models/wide_resnet50_2-95faca4d.pth",
    "wide_resnet101_2": "https://download.pytorch.org/models/wide_resnet101_2-32ee1156.pth",
}


def load_torch_weights(
    model: eqx.Module, filepath: Path = None, url: "_Url" = None
) -> eqx.Module:
    """Loads weights from a PyTorch serialised file.

    ???+ warning

        This method requires installation of the [`torch`](https://pypi.org/project/torch/) package.

    !!! note

        - This function assumes that Eqxvision's ordering of class
          attributes mirrors the `torchvision.models` implementation.
        - The saved checkpoint should **only** contain model parameters as keys.

    **Arguments:**

    - model: An `eqx.Module` for which the `jnp.ndarray` leaves are
        replaced by corresponding `PyTorch` weights.
    - filepath: `Path` to the downloaded `PyTorch` model file.
    - url: `URL` for the `PyTorch` model file. The file is downloaded to `/tmp/.eqx/` folder.

    **Returns:**
        The model with weights loaded from the `PyTorch` checkpoint.
    """
    if "torch" not in sys.modules:
        raise RuntimeError(
            " Torch package not found! Pretrained is only supported with the torch package."
        )

    if filepath is None and url is None:
        raise ValueError("Both filepath and url cannot be empty!")
    elif filepath and url:
        warnings.warn(f"Overriding `url` with with filepath: {filepath}.")
        url = None
    if url:
        global _TEMP_DIR
        filepath = os.path.join(_TEMP_DIR, os.path.basename(url))
        if os.path.exists(filepath):
            logging.info(
                f"Downloaded file exists at f{filepath}. Using the cached file!"
            )
        else:
            os.makedirs(_TEMP_DIR, exist_ok=True)
            torch.hub.download_url_to_file(url, filepath)
    if not os.path.exists(filepath):
        raise ValueError(f"filepath: {filepath} does not exist!")

    weights = torch.load(filepath, map_location="cpu")
    weights_iterator = iter(
        [
            (name, jnp.asarray(weight.detach().numpy()))
            for name, weight in weights.items()
            if "running" not in name and "num_batches" not in name
        ]
    )

    bn_s = []
    for name, weight in weights.items():
        if "running_mean" in name:
            bn_s.append(False)
            bn_s.append(jnp.asarray(weight.detach().numpy()))
        elif "running_var" in name:
            bn_s.append(jnp.asarray(weight.detach().numpy()))
    bn_iterator = iter(bn_s)

    leaves, tree_def = jtu.tree_flatten(model)

    new_leaves = []
    for leaf in leaves:
        if isinstance(leaf, jnp.ndarray) and not (
            leaf.size == 1 and isinstance(leaf.item(), bool)
        ):
            (weight_name, new_weights) = next(weights_iterator)
            new_leaves.append(jnp.reshape(new_weights, leaf.shape))
        else:
            new_leaves.append(leaf)

    model = jtu.tree_unflatten(tree_def, new_leaves)

    def set_experimental(iter_bn, x):
        def set_values(y):
            if isinstance(y, eqx.experimental.StateIndex):
                current_val = next(iter_bn)
                if isinstance(current_val, bool):
                    eqx.experimental.set_state(y, jnp.asarray(False))
                else:
                    running_mean, running_var = current_val, next(iter_bn)
                    eqx.experimental.set_state(y, (running_mean, running_var))
            return y

        return jtu.tree_map(
            set_values, x, is_leaf=lambda _: isinstance(_, eqx.experimental.StateIndex)
        )

    model = jtu.tree_map(set_experimental, bn_iterator, model)
    return model


def _make_divisible(v: float, divisor: int, min_value: Optional[int] = None) -> int:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v



from typing import Any, cast, Dict, List, Optional, Union

import equinox as eqx
import equinox.experimental as eqex
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...utils import load_torch_weights, MODEL_URLS


_cfgs: Dict[str, List[Union[str, int]]] = {
    "A": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "B": [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "D": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        "M",
    ],
    "E": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        512,
        "M",
    ],
}


class VGG(eqx.Module):
    """A simple port of `torchvision.models.vgg`"""

    features: nn.Sequential
    avgpool: nn.AdaptiveAvgPool2d
    classifier: nn.Sequential

    def __init__(
        self,
        cfg: List[Union[str, int]] = None,
        num_classes: int = 1000,
        batch_norm: bool = True,
        dropout: float = 0.5,
        *,
        key: Optional["jax.random.PRNGKey"] = None
    ) -> None:
        """**Arguments:**

        - `cfg`: A list specifying the block configuration.
        - `num_classes`: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`.
        - `batch_norm` : If `True`, then `BatchNorm` is enabled in the architecture.
        - `dropout`: The probability parameter for `equinox.nn.Dropout`.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
        """
        super(VGG, self).__init__()
        if key is None:
            key = jrandom.PRNGKey(0)
        keys = jrandom.split(key, 4)

        self.features = _make_layers(cfg, batch_norm, key=keys[0])
        self.avgpool = nn.AdaptiveAvgPool2d((7, 7))
        self.classifier = nn.Sequential(
            [
                nn.Linear(512 * 7 * 7, 4096, key=keys[1]),
                nn.Dropout(p=dropout),
                nn.Linear(4096, 4096, key=keys[2]),
                nn.Lambda(jnn.relu),
                nn.Dropout(p=dropout),
                nn.Linear(4096, num_classes, key=keys[3]),
            ]
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array with `3` channels.
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`.
        """
        keys = jrandom.split(key, 2)
        x = self.features(x, key=keys[0])
        x = self.avgpool(x)
        x = jnp.ravel(x)
        x = self.classifier(x, key=keys[1])
        return x


def _make_layers(
    cfg: List[Union[str, int]],
    batch_norm: bool = False,
    key: "jax.random.PRNGKey" = None,
) -> nn.Sequential:

    layers: List[eqx.Module] = []
    in_channels = 3
    keys = jrandom.split(key=key, num=len(cfg) - cfg.count("M"))
    count = 0
    for v in cfg:
        if v == "M":
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
        else:
            v = cast(int, v)
            conv2d = nn.Conv2d(
                in_channels, v, kernel_size=3, padding=1, key=keys[count]
            )
            if batch_norm:
                layers += [
                    conv2d,
                    eqex.BatchNorm(v, axis_name="batch"),
                    nn.Lambda(jnn.relu),
                ]
            else:
                layers += [conv2d, nn.Lambda(jnn.relu)]
            in_channels = v
            count += 1
    return nn.Sequential(layers)


def _vgg(cfg: str, batch_norm: bool, **kwargs: Any) -> VGG:
    model = VGG(cfg=_cfgs[cfg], batch_norm=batch_norm, **kwargs)
    return model


def vgg11(pretrained=False, **kwargs: Any) -> VGG:
    r"""VGG 11-layer model (configuration "A") from
    [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556.pdf).
    The required minimum input size of the model is 32x32.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _vgg("A", False, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["vgg11"])
    return model


def vgg11_bn(pretrained=False, **kwargs: Any) -> VGG:
    r"""VGG 11-layer model (configuration "A") with batch normalization
    [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556.pdf).
    The required minimum input size of the model is 32x32.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _vgg("A", True, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["vgg11_bn"])
    return model


def vgg13(pretrained=False, **kwargs: Any) -> VGG:
    r"""VGG 13-layer model (configuration "B")
    [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556.pdf).
    The required minimum input size of the model is 32x32.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _vgg("B", False, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["vgg13"])
    return model


def vgg13_bn(pretrained=False, **kwargs: Any) -> VGG:
    r"""VGG 13-layer model (configuration "B") with batch normalization
    [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556.pdf).
    The required minimum input size of the model is 32x32.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _vgg("B", True, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["vgg13_bn"])
    return model


def vgg16(pretrained=False, **kwargs: Any) -> VGG:
    r"""VGG 16-layer model (configuration "D")
    [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556.pdf).
    The required minimum input size of the model is 32x32.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _vgg("D", False, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["vgg16"])
    return model


def vgg16_bn(pretrained=False, **kwargs: Any) -> VGG:
    r"""VGG 16-layer model (configuration "D") with batch normalization
    [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556.pdf).
    The required minimum input size of the model is 32x32.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _vgg("D", True, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["vgg16_bn"])
    return model


def vgg19(pretrained=False, **kwargs: Any) -> VGG:
    r"""VGG 19-layer model (configuration "E")
    [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556.pdf).
    The required minimum input size of the model is 32x32.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _vgg("E", False, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["vgg19"])
    return model


def vgg19_bn(pretrained=False, **kwargs: Any) -> VGG:
    r"""VGG 19-layer model (configuration 'E') with batch normalization
    [Very Deep Convolutional Networks For Large-Scale Image Recognition](https://arxiv.org/pdf/1409.1556.pdf).
    The required minimum input size of the model is 32x32.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _vgg("E", True, **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["vgg19_bn"])
    return model



from . import (
    alexnet,
    convnext,
    densenet,
    efficientnet,
    googlenet,
    mobilenetv2,
    mobilenetv3,
    regnet,
    resnet,
    shufflenetv2,
    squeezenet,
    swin,
    vgg,
    vit,
)



from .conv_norm_activation import ConvNormActivation
from .drop_path import DropPath
from .extensions_2d import LayerNorm2d, Linear2d
from .mlps import MlpProjection
from .patch_embed import PatchEmbed
from .squeeze import SqueezeExcitation



import copy
import math
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, List, Optional, Sequence, Tuple, Union

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jr
from equinox.custom_types import Array

from ...layers import ConvNormActivation, DropPath, SqueezeExcitation
from ...utils import _make_divisible, load_torch_weights, MODEL_URLS


@dataclass
class _MBConvConfigData:
    expand_ratio: float
    kernel: int
    stride: int
    input_channels: int
    out_channels: int
    num_layers: int
    block: Callable[..., eqx.Module]

    @staticmethod
    def adjust_channels(
        channels: int, width_mult: float, min_value: Optional[int] = None
    ) -> int:
        return _make_divisible(channels * width_mult, 8, min_value)


class _MBConvConfig(_MBConvConfigData):
    # Stores information listed at Table 1 of the EfficientNet paper & Table 4 of the EfficientNetV2 paper
    def __init__(
        self,
        expand_ratio: float,
        kernel: int,
        stride: int,
        input_channels: int,
        out_channels: int,
        num_layers: int,
        width_mult: float = 1.0,
        depth_mult: float = 1.0,
        block: Optional[Callable[..., eqx.Module]] = None,
    ) -> None:
        input_channels = self.adjust_channels(input_channels, width_mult)
        out_channels = self.adjust_channels(out_channels, width_mult)
        num_layers = self.adjust_depth(num_layers, depth_mult)
        if block is None:
            block = _MBConv
        super().__init__(
            expand_ratio,
            kernel,
            stride,
            input_channels,
            out_channels,
            num_layers,
            block,
        )

    @staticmethod
    def adjust_depth(num_layers: int, depth_mult: float):
        return int(math.ceil(num_layers * depth_mult))


class _FusedMBConvConfig(_MBConvConfigData):
    # Stores information listed at Table 4 of the EfficientNetV2 paper
    def __init__(
        self,
        expand_ratio: float,
        kernel: int,
        stride: int,
        input_channels: int,
        out_channels: int,
        num_layers: int,
        block: Optional[Callable[..., eqx.Module]] = None,
    ) -> None:
        if block is None:
            block = _FusedMBConv
        super().__init__(
            expand_ratio,
            kernel,
            stride,
            input_channels,
            out_channels,
            num_layers,
            block,
        )


class _MBConv(eqx.Module):
    use_res_connect: bool
    block: nn.Sequential
    stochastic_depth: DropPath
    out_channels: int

    def __init__(
        self,
        cnf: _MBConvConfig,
        stochastic_depth_prob: float,
        norm_layer: Callable[..., eqx.Module],
        se_layer: Callable[..., eqx.Module] = SqueezeExcitation,
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        super().__init__()

        if not (1 <= cnf.stride <= 2):
            raise ValueError("illegal stride value")

        if key is None:
            key = jr.PRNGKey(0)
        keys = jr.split(key, 4)
        self.use_res_connect = (
            cnf.stride == 1 and cnf.input_channels == cnf.out_channels
        )

        layers: List[eqx.Module] = []
        activation_layer = jnn.silu

        # expand
        expanded_channels = cnf.adjust_channels(cnf.input_channels, cnf.expand_ratio)
        if expanded_channels != cnf.input_channels:
            layers.append(
                ConvNormActivation(
                    cnf.input_channels,
                    expanded_channels,
                    kernel_size=1,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                    key=keys[0],
                )
            )

        # depthwise
        layers.append(
            ConvNormActivation(
                expanded_channels,
                expanded_channels,
                kernel_size=cnf.kernel,
                stride=cnf.stride,
                groups=expanded_channels,
                norm_layer=norm_layer,
                activation_layer=activation_layer,
                key=keys[1],
            )
        )

        # squeeze and excitation
        squeeze_channels = max(1, cnf.input_channels // 4)
        layers.append(
            se_layer(
                expanded_channels,
                squeeze_channels,
                activation=activation_layer,
                key=keys[2],
            )
        )

        # project
        layers.append(
            ConvNormActivation(
                expanded_channels,
                cnf.out_channels,
                kernel_size=1,
                norm_layer=norm_layer,
                activation_layer=None,
                key=keys[3],
            )
        )

        self.block = nn.Sequential(layers)
        self.stochastic_depth = DropPath(stochastic_depth_prob, mode="per_channel")
        self.out_channels = cnf.out_channels

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        keys = jr.split(key, 2)
        result = self.block(x, key=keys[0])
        if self.use_res_connect:
            result = self.stochastic_depth(result, key=keys[1])
            result += x
        return result


class _FusedMBConv(eqx.Module):
    use_res_connect: bool
    block: nn.Sequential
    stochastic_depth: DropPath
    out_channels: int

    def __init__(
        self,
        cnf: _FusedMBConvConfig,
        stochastic_depth_prob: float,
        norm_layer: Callable[..., eqx.Module],
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        super().__init__()

        if not (1 <= cnf.stride <= 2):
            raise ValueError("illegal stride value")
        if key is None:
            key = jr.PRNGKey(0)
        keys = jr.split(key, 3)
        self.use_res_connect = (
            cnf.stride == 1 and cnf.input_channels == cnf.out_channels
        )

        layers: List[eqx.Module] = []
        activation_layer = jnn.silu

        expanded_channels = cnf.adjust_channels(cnf.input_channels, cnf.expand_ratio)
        if expanded_channels != cnf.input_channels:
            # fused expand
            layers.append(
                ConvNormActivation(
                    cnf.input_channels,
                    expanded_channels,
                    kernel_size=cnf.kernel,
                    stride=cnf.stride,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                    key=keys[0],
                )
            )

            # project
            layers.append(
                ConvNormActivation(
                    expanded_channels,
                    cnf.out_channels,
                    kernel_size=1,
                    norm_layer=norm_layer,
                    activation_layer=None,
                    key=keys[1],
                )
            )
        else:
            layers.append(
                ConvNormActivation(
                    cnf.input_channels,
                    cnf.out_channels,
                    kernel_size=cnf.kernel,
                    stride=cnf.stride,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                    key=keys[2],
                )
            )

        self.block = nn.Sequential(layers)
        self.stochastic_depth = DropPath(stochastic_depth_prob, mode="local")
        self.out_channels = cnf.out_channels

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        keys = jr.split(key, 2)
        result = self.block(x, key=keys[0])
        if self.use_res_connect:
            result = self.stochastic_depth(result, key=keys[1])
            result += x
        return result


class EfficientNet(eqx.Module):
    """A simple port of `torchvision.models.efficientnet`."""

    features: nn.Sequential
    avgpool: nn.AdaptiveAvgPool2d
    classifier: nn.Sequential

    def __init__(
        self,
        inverted_residual_setting: Sequence[
            Union["_MBConvConfig", "_FusedMBConvConfig"]
        ],
        dropout: float,
        stochastic_depth_prob: float = 0.2,
        num_classes: int = 1000,
        norm_layer: Optional["eqx.Module"] = None,
        last_channel: Optional[int] = None,
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        """**Arguments:**

        - `inverted_residual_setting`: Network structure
        - `dropout`: The dropout probability
        - `stochastic_depth_prob`: Probability of dropping a sample along channels
        - `num_classes`: Number of classes in the classification task.
                         Also controls the final output shape `(num_classes,)`
        - `norm_layer`: Normalisation applied to the intermediate outputs
        - `last_channel`: The number of channels on the penultimate layer
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)

        """
        super().__init__()

        if not inverted_residual_setting:
            raise ValueError("The inverted_residual_setting should not be empty")
        elif not (
            isinstance(inverted_residual_setting, Sequence)
            and all(
                [isinstance(s, _MBConvConfigData) for s in inverted_residual_setting]
            )
        ):
            raise TypeError(
                "The inverted_residual_setting should be List[MBConvConfig]"
            )

        if key is None:
            key = jr.PRNGKey(0)
        keys = jr.split(key, 3)

        if norm_layer is None:
            norm_layer = eqx.experimental.BatchNorm

        layers: List[eqx.Module] = []

        # building first layer
        firstconv_output_channels = inverted_residual_setting[0].input_channels
        layers.append(
            ConvNormActivation(
                3,
                firstconv_output_channels,
                kernel_size=3,
                stride=2,
                norm_layer=norm_layer,
                activation_layer=jnn.silu,
                key=keys[0],
            )
        )

        # building inverted residual blocks
        total_stage_blocks = sum(cnf.num_layers for cnf in inverted_residual_setting)
        stage_block_id = 0
        for cnf in inverted_residual_setting:
            stage: List[eqx.Module] = []
            for _ in range(cnf.num_layers):
                keys = jr.split(keys[1], 2)
                # copy to avoid modifications. shallow copy is enough
                block_cnf = copy.copy(cnf)

                # overwrite info if not the first conv in the stage
                if stage:
                    block_cnf.input_channels = block_cnf.out_channels
                    block_cnf.stride = 1

                # adjust stochastic depth probability based on the depth of the stage block
                sd_prob = (
                    stochastic_depth_prob * float(stage_block_id) / total_stage_blocks
                )

                stage.append(
                    block_cnf.block(block_cnf, sd_prob, norm_layer, key=keys[0])
                )
                stage_block_id += 1

            layers.append(nn.Sequential(stage))

        keys = jr.split(keys[1], 2)
        # building last several layers
        lastconv_input_channels = inverted_residual_setting[-1].out_channels
        lastconv_output_channels = (
            last_channel if last_channel is not None else 4 * lastconv_input_channels
        )
        layers.append(
            ConvNormActivation(
                lastconv_input_channels,
                lastconv_output_channels,
                kernel_size=1,
                norm_layer=norm_layer,
                activation_layer=jnn.silu,
                key=keys[0],
            )
        )

        self.features = nn.Sequential(layers)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.classifier = nn.Sequential(
            [
                nn.Dropout(p=dropout),
                nn.Linear(lastconv_output_channels, num_classes, key=keys[1]),
            ]
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array.
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`.
        """
        keys = jr.split(key, 2)
        x = self.features(x, key=keys[0])
        x = self.avgpool(x)
        x = jnp.ravel(x)
        x = self.classifier(x, key=keys[1])
        return x


def _efficientnet(
    arch: str,
    inverted_residual_setting: Sequence[Union[_MBConvConfig, _FusedMBConvConfig]],
    dropout: float,
    last_channel: Optional[int],
    pretrained: bool,
    **kwargs: Any,
) -> EfficientNet:
    model = EfficientNet(
        inverted_residual_setting, dropout, last_channel=last_channel, **kwargs
    )
    if pretrained:
        if arch not in MODEL_URLS:
            raise ValueError(f"No checkpoint is available for model type {arch}")
        model = load_torch_weights(model, url=MODEL_URLS[arch])

    return model


def _efficientnet_conf(
    arch: str,
    **kwargs: Any,
) -> Tuple[Sequence[Union[_MBConvConfig, _FusedMBConvConfig]], Optional[int]]:
    inverted_residual_setting: Sequence[Union[_MBConvConfig, _FusedMBConvConfig]]
    if arch.startswith("efficientnet_b"):
        bneck_conf = partial(
            _MBConvConfig,
            width_mult=kwargs.pop("width_mult"),
            depth_mult=kwargs.pop("depth_mult"),
        )
        inverted_residual_setting = [
            bneck_conf(1, 3, 1, 32, 16, 1),
            bneck_conf(6, 3, 2, 16, 24, 2),
            bneck_conf(6, 5, 2, 24, 40, 2),
            bneck_conf(6, 3, 2, 40, 80, 3),
            bneck_conf(6, 5, 1, 80, 112, 3),
            bneck_conf(6, 5, 2, 112, 192, 4),
            bneck_conf(6, 3, 1, 192, 320, 1),
        ]
        last_channel = None
    elif arch.startswith("efficientnet_v2_s"):
        inverted_residual_setting = [
            _FusedMBConvConfig(1, 3, 1, 24, 24, 2),
            _FusedMBConvConfig(4, 3, 2, 24, 48, 4),
            _FusedMBConvConfig(4, 3, 2, 48, 64, 4),
            _MBConvConfig(4, 3, 2, 64, 128, 6),
            _MBConvConfig(6, 3, 1, 128, 160, 9),
            _MBConvConfig(6, 3, 2, 160, 256, 15),
        ]
        last_channel = 1280
    elif arch.startswith("efficientnet_v2_m"):
        inverted_residual_setting = [
            _FusedMBConvConfig(1, 3, 1, 24, 24, 3),
            _FusedMBConvConfig(4, 3, 2, 24, 48, 5),
            _FusedMBConvConfig(4, 3, 2, 48, 80, 5),
            _MBConvConfig(4, 3, 2, 80, 160, 7),
            _MBConvConfig(6, 3, 1, 160, 176, 14),
            _MBConvConfig(6, 3, 2, 176, 304, 18),
            _MBConvConfig(6, 3, 1, 304, 512, 5),
        ]
        last_channel = 1280
    elif arch.startswith("efficientnet_v2_l"):
        inverted_residual_setting = [
            _FusedMBConvConfig(1, 3, 1, 32, 32, 4),
            _FusedMBConvConfig(4, 3, 2, 32, 64, 7),
            _FusedMBConvConfig(4, 3, 2, 64, 96, 7),
            _MBConvConfig(4, 3, 2, 96, 192, 10),
            _MBConvConfig(6, 3, 1, 192, 224, 19),
            _MBConvConfig(6, 3, 2, 224, 384, 25),
            _MBConvConfig(6, 3, 1, 384, 640, 7),
        ]
        last_channel = 1280
    else:
        raise ValueError(f"Unsupported model type {arch}")

    return inverted_residual_setting, last_channel


def efficientnet_b0(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """EfficientNet B0 model architecture from the [EfficientNet: Rethinking Model Scaling for Convolutional
    Neural Networks](https://arxiv.org/abs/1905.11946) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b0", width_mult=1.0, depth_mult=1.0
    )
    return _efficientnet(
        "efficientnet_b0",
        inverted_residual_setting,
        0.2,
        last_channel,
        pretrained,
        **kwargs,
    )


def efficientnet_b1(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """EfficientNet B1 model architecture from the [EfficientNet: Rethinking Model Scaling for Convolutional
    Neural Networks](https://arxiv.org/abs/1905.11946) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """

    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b1", width_mult=1.0, depth_mult=1.1
    )
    return _efficientnet(
        "efficientnet_b1",
        inverted_residual_setting,
        0.2,
        last_channel,
        pretrained,
        **kwargs,
    )


def efficientnet_b2(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """EfficientNet B2 model architecture from the [EfficientNet: Rethinking Model Scaling for Convolutional
    Neural Networks](https://arxiv.org/abs/1905.11946) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """

    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b2", width_mult=1.1, depth_mult=1.2
    )
    return _efficientnet(
        "efficientnet_b2",
        inverted_residual_setting,
        0.3,
        last_channel,
        pretrained,
        **kwargs,
    )


def efficientnet_b3(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """EfficientNet B3 model architecture from the [EfficientNet: Rethinking Model Scaling for Convolutional
    Neural Networks](https://arxiv.org/abs/1905.11946) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b3", width_mult=1.2, depth_mult=1.4
    )
    return _efficientnet(
        "efficientnet_b3",
        inverted_residual_setting,
        0.3,
        last_channel,
        pretrained,
        **kwargs,
    )


def efficientnet_b4(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """EfficientNet B4 model architecture from the [EfficientNet: Rethinking Model Scaling for Convolutional
    Neural Networks](https://arxiv.org/abs/1905.11946) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b4", width_mult=1.4, depth_mult=1.8
    )
    return _efficientnet(
        "efficientnet_b4",
        inverted_residual_setting,
        0.4,
        last_channel,
        pretrained,
        **kwargs,
    )


def efficientnet_b5(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """EfficientNet B5 model architecture from the [EfficientNet: Rethinking Model Scaling for Convolutional
    Neural Networks](https://arxiv.org/abs/1905.11946) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b5", width_mult=1.6, depth_mult=2.2
    )
    return _efficientnet(
        "efficientnet_b5",
        inverted_residual_setting,
        0.4,
        last_channel,
        pretrained,
        norm_layer=partial(eqx.experimental.BatchNorm, eps=0.001, momentum=0.01),
        **kwargs,
    )


def efficientnet_b6(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """EfficientNet B6 model architecture from the [EfficientNet: Rethinking Model Scaling for Convolutional
    Neural Networks](https://arxiv.org/abs/1905.11946) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b6", width_mult=1.8, depth_mult=2.6
    )
    return _efficientnet(
        "efficientnet_b6",
        inverted_residual_setting,
        0.5,
        last_channel,
        pretrained,
        norm_layer=partial(eqx.experimental.BatchNorm, eps=0.001, momentum=0.01),
        **kwargs,
    )


def efficientnet_b7(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """EfficientNet B7 model architecture from the [EfficientNet: Rethinking Model Scaling for Convolutional
    Neural Networks](https://arxiv.org/abs/1905.11946) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b7", width_mult=2.0, depth_mult=3.1
    )
    return _efficientnet(
        "efficientnet_b7",
        inverted_residual_setting,
        0.5,
        last_channel,
        pretrained,
        norm_layer=partial(eqx.experimental.BatchNorm, eps=0.001, momentum=0.01),
        **kwargs,
    )


def efficientnet_v2_s(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """
    Constructs an EfficientNetV2-S architecture from
    [EfficientNetV2: Smaller Models and Faster Training](https://arxiv.org/abs/2104.00298).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_s")
    return _efficientnet(
        "efficientnet_v2_s",
        inverted_residual_setting,
        0.2,
        last_channel,
        pretrained,
        norm_layer=partial(eqx.experimental.BatchNorm, eps=1e-03),
        **kwargs,
    )


def efficientnet_v2_m(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """
    Constructs an EfficientNetV2-M architecture from
    [EfficientNetV2: Smaller Models and Faster Training](https://arxiv.org/abs/2104.00298).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_m")
    return _efficientnet(
        "efficnetnet_v2_m",
        inverted_residual_setting,
        0.3,
        last_channel,
        pretrained,
        norm_layer=partial(eqx.experimental.BatchNorm, eps=1e-03),
        **kwargs,
    )


def efficientnet_v2_l(pretrained: bool = False, **kwargs: Any) -> EfficientNet:
    """
    Constructs an EfficientNetV2-L architecture from
    [EfficientNetV2: Smaller Models and Faster Training](https://arxiv.org/abs/2104.00298).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_l")
    return _efficientnet(
        "efficientnet_v2_l",
        inverted_residual_setting,
        0.4,
        last_channel,
        pretrained,
        norm_layer=partial(eqx.experimental.BatchNorm, eps=1e-03),
        **kwargs,
    )



from typing import Callable, Tuple, Union

import equinox as eqx
import equinox.nn as nn
import jax
import jax.random as jrandom
from equinox.custom_types import Array

from .extensions_2d import Linear2d


class MlpProjection(eqx.Module):
    """MLP as used in Vision Transformer, MLP-Mixer and related networks"""

    fc1: eqx.Module
    act: Callable
    drop1: nn.Dropout
    fc2: eqx.Module
    drop2: nn.Dropout

    def __init__(
        self,
        in_features: int,
        hidden_features: int = None,
        out_features: int = None,
        lin_layer: Union["Linear2d", "nn.Linear"] = nn.Linear,
        act_layer: Callable = None,
        drop: Union[float, Tuple[float]] = 0.0,
        *,
        key: "jax.random.PRNGKey" = None
    ):
        """**Arguments:**

        - `in_features`: The expected dimension of the input
        - `hidden_features`: Dimensionality of the hidden layer
        - `out_features`: The dimension of the output feature
        - `lin_layer`: Linear layer to use. For transformer like architectures, `Linear2d` can be easier to integrate.
        - `act_layer`: Activation function to be applied to the intermediate layers
        - `drop`: The probability associated with `Dropout`
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
        """
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        drop_probs = drop if isinstance(drop, tuple) else (drop, drop)
        keys = jrandom.split(key, 2)
        self.fc1 = lin_layer(in_features, hidden_features, key=keys[0])
        self.act = act_layer
        self.drop1 = nn.Dropout(drop_probs[0])
        self.fc2 = lin_layer(hidden_features, out_features, key=keys[1])
        self.drop2 = nn.Dropout(drop_probs[1])

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Utilised by few layers in the network such as `Dropout` or `DropPath`
        """
        keys = jrandom.split(key, 2)
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x, key=keys[0])
        x = self.fc2(x)
        x = self.drop2(x, key=keys[1])
        return x



from typing import Optional

import equinox as eqx
import jax
import jax.numpy as jnp
from equinox.custom_types import Array


class LayerNorm2d(eqx.nn.LayerNorm):
    """Wraps `eqx.nn.LayerNorm` for an easy to apply channelwise-variant."""

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array of shape `(channels, dim_0, dim_1)`
        - `key`: Ignored

        **Returns:**

        Output of `eqx.nn.LayerNorm` applied to each `dim_0*dim_1 x c` entry.
        """
        c, h, w = x.shape
        x = jnp.transpose(x.reshape(c, -1))
        x = jax.vmap(super(LayerNorm2d, self).__call__)(x)
        x = jnp.transpose(x).reshape(c, h, w)
        return x


class Linear2d(eqx.nn.Linear):
    """Wraps `eqx.nn.Linear` for an easy to apply channelwise-variant."""

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array of shape `(channels, dim_0, dim_1)`
        - `key`: Ignored

        **Returns:**

        Output of `eqx.nn.Linear` applied to each `dim_0*dim_1 x c` entry.
        """
        c, h, w = x.shape
        x = jnp.transpose(x.reshape(c, -1))
        x = jax.vmap(super(Linear2d, self).__call__)(x)
        x = jnp.transpose(x).reshape(self.out_features, h, w)
        return x



from typing import Callable, Optional

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.random as jrandom
from equinox.custom_types import Array


class SqueezeExcitation(eqx.Module):
    """A simple port of `torchvision.ops.misc.SqueezeExcitation`"""

    avgpool: nn.AdaptiveAvgPool2d
    fc1: nn.Conv2d
    fc2: nn.Conv2d
    activation: nn.Lambda
    scale_activation: nn.Lambda

    def __init__(
        self,
        input_channels: int,
        squeeze_channels: int,
        activation: Callable = None,
        scale_activation: Callable = None,
        *,
        key: "jax.random.PRNGKey" = None
    ) -> None:
        """**Arguments:**

        - `input_channels`: Number of channels in the input image
        - `squeeze_channels`: Number of squeeze channels
        - `activation`: ``delta`` activation. Defaults to `jax.nn.relu`
        - `scale_activation`: `sigma`` activation. Defaults to `jax.nn.sigmoid`
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        super().__init__()
        if activation is None:
            activation = jnn.relu
        if scale_activation is None:
            scale_activation = jnn.sigmoid

        keys = jrandom.split(key, 2)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.fc1 = nn.Conv2d(input_channels, squeeze_channels, 1, key=keys[0])
        self.fc2 = nn.Conv2d(squeeze_channels, input_channels, 1, key=keys[1])
        self.activation = nn.Lambda(activation)
        self.scale_activation = nn.Lambda(scale_activation)

    def __call__(self, x, *, key: Optional["jax.random.PRNGKey"] = None) -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Forwarded to individual `eqx.Module` attributes
        """
        scale = self.avgpool(x)
        scale = self.fc1(scale)
        scale = self.activation(scale, key=key)
        scale = self.fc2(scale)
        return x * self.scale_activation(scale, key=key)



from typing import Any, List, Optional

import equinox as eqx
import equinox.experimental as eqxex
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...utils import load_torch_weights, MODEL_URLS


def _channel_shuffle(x: Array, groups: int) -> Array:
    num_channels, height, width = x.shape
    channels_per_group = num_channels // groups
    x = jnp.reshape(x, (groups, channels_per_group, height, width))
    x = jnp.transpose(x, axes=(1, 0, 2, 3))
    x = jnp.reshape(x, (-1, height, width))
    return x


class _InvertedResidual(eqx.Module):
    stride: int
    branch1: nn.Sequential
    branch2: nn.Sequential

    def __init__(
        self,
        inp: int,
        oup: int,
        stride: int,
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        super().__init__()

        keys = jrandom.split(key, 5)

        if not (1 <= stride <= 3):
            raise ValueError("illegal stride value")

        branch_features = oup // 2
        assert (stride != 1) or (inp == branch_features << 1)

        self.stride = stride
        if stride > 1:
            self.branch1 = nn.Sequential(
                [
                    self.depthwise_conv(
                        inp,
                        inp,
                        kernel_size=3,
                        stride=self.stride,
                        padding=1,
                        key=keys[0],
                    ),
                    eqxex.BatchNorm(inp, axis_name="batch"),
                    nn.Conv2d(
                        inp,
                        branch_features,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        use_bias=False,
                        key=keys[1],
                    ),
                    eqxex.BatchNorm(branch_features, axis_name="batch"),
                    nn.Lambda(jnn.relu),
                ]
            )
        else:
            self.branch1 = nn.Sequential([nn.Identity])

        self.branch2 = nn.Sequential(
            [
                nn.Conv2d(
                    inp if (self.stride > 1) else branch_features,
                    branch_features,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    use_bias=False,
                    key=keys[2],
                ),
                eqxex.BatchNorm(branch_features, axis_name="batch"),
                nn.Lambda(jnn.relu),
                self.depthwise_conv(
                    branch_features,
                    branch_features,
                    kernel_size=3,
                    stride=self.stride,
                    padding=1,
                    key=keys[3],
                ),
                eqxex.BatchNorm(branch_features, axis_name="batch"),
                nn.Conv2d(
                    branch_features,
                    branch_features,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    use_bias=False,
                    key=keys[4],
                ),
                eqxex.BatchNorm(branch_features, axis_name="batch"),
                nn.Lambda(jnn.relu),
            ]
        )

    @staticmethod
    def depthwise_conv(
        i: int,
        o: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 0,
        bias: bool = False,
        key=None,
    ) -> nn.Conv2d:
        return nn.Conv2d(
            i, o, kernel_size, stride, padding, use_bias=bias, groups=i, key=key
        )

    def __call__(self, x, *, key: "jax.random.PRNGKey") -> Array:
        if self.stride == 1:
            x1, x2 = jnp.split(x, 2, axis=0)
            out = jnp.concatenate((x1, self.branch2(x2)), axis=0)
        else:
            out = jnp.concatenate((self.branch1(x), self.branch2(x)), axis=0)

        out = _channel_shuffle(out, 2)
        return out


class ShuffleNetV2(eqx.Module):
    """A simple port of `torchvision.models.shufflenetv2`"""

    conv1: nn.Sequential
    maxpool: nn.MaxPool2d
    stage2: nn.Sequential
    stage3: nn.Sequential
    stage4: nn.Sequential
    conv5: nn.Sequential
    pool: nn.AdaptiveAvgPool2d
    fc: nn.Linear

    def __init__(
        self,
        stages_repeats: List[int],
        stages_out_channels: List[int],
        num_classes: int = 1000,
        inverted_residual: "eqx.Module" = _InvertedResidual,
        *,
        key: Optional["jax.random.PRNGKey"] = None,
    ) -> None:
        """**Arguments:**

        - stages_repeats: Number of times a block is repeated for each stage
        - stages_out_channels: Output at each stage
        - num_classes: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        - inverted_residual: Network structure
        - key: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        super().__init__()
        if key is None:
            key = jrandom.PRNGKey(0)
        keys = jrandom.split(key, 2)

        if len(stages_repeats) != 3:
            raise ValueError("expected stages_repeats as list of 3 positive ints")
        if len(stages_out_channels) != 5:
            raise ValueError("expected stages_out_channels as list of 5 positive ints")

        input_channels = 3
        output_channels = stages_out_channels[0]
        self.conv1 = nn.Sequential(
            [
                nn.Conv2d(
                    input_channels,
                    output_channels,
                    3,
                    2,
                    1,
                    use_bias=False,
                    key=keys[0],
                ),
                eqxex.BatchNorm(output_channels, axis_name="batch"),
                nn.Lambda(jnn.relu),
            ]
        )
        input_channels = output_channels

        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        # Static annotations for mypy
        self.stage2: nn.Sequential
        self.stage3: nn.Sequential
        self.stage4: nn.Sequential
        stage_names = [f"stage{i}" for i in [2, 3, 4]]
        for name, repeats, output_channels in zip(
            stage_names, stages_repeats, stages_out_channels[1:]
        ):
            keys = jrandom.split(keys[1], 2)
            seq = [inverted_residual(input_channels, output_channels, 2, key=keys[0])]
            for i in range(repeats - 1):
                keys = jrandom.split(keys[1], 2)
                seq.append(
                    inverted_residual(output_channels, output_channels, 1, key=keys[0])
                )
            setattr(self, name, nn.Sequential(seq))
            input_channels = output_channels

        keys = jrandom.split(keys[1], 2)
        output_channels = stages_out_channels[-1]
        self.conv5 = nn.Sequential(
            [
                nn.Conv2d(
                    input_channels,
                    output_channels,
                    1,
                    1,
                    0,
                    use_bias=False,
                    key=keys[0],
                ),
                eqxex.BatchNorm(output_channels, axis_name="batch"),
                nn.Lambda(jnn.relu),
            ]
        )
        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(output_channels, num_classes, key=keys[1])

    def __call__(self, x, *, key: Optional["jax.random.PRNGKey"] = None) -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        keys = jrandom.split(key, 5)
        x = self.conv1(x, key=keys[0])
        x = self.maxpool(x)
        x = self.stage2(x, key=keys[1])
        x = self.stage3(x, key=keys[2])
        x = self.stage4(x, key=keys[3])
        x = self.conv5(x, key=keys[4])
        x = jnp.ravel(self.pool(x))
        x = self.fc(x)
        return x


def _shufflenetv2(*args: Any, **kwargs: Any) -> ShuffleNetV2:
    model = ShuffleNetV2(*args, **kwargs)
    return model


def shufflenet_v2_x0_5(pretrained=False, **kwargs: Any) -> ShuffleNetV2:
    """
    Constructs a ShuffleNetV2 with 0.5x output channels, as described in
    [ShuffleNet V2: Practical Guidelines for Efficient CNN Architecture Design](https://arxiv.org/abs/1807.11164).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _shufflenetv2([4, 8, 4], [24, 48, 96, 192, 1024], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["shufflenetv2_x0.5"])
    return model


def shufflenet_v2_x1_0(pretrained=False, **kwargs: Any) -> ShuffleNetV2:
    """
    Constructs a ShuffleNetV2 with 1.0x output channels, as described in
    [ShuffleNet V2: Practical Guidelines for Efficient CNN Architecture Design](https://arxiv.org/abs/1807.11164).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _shufflenetv2([4, 8, 4], [24, 116, 232, 464, 1024], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["shufflenetv2_x1.0"])
    return model


def shufflenet_v2_x1_5(pretrained=False, **kwargs: Any) -> ShuffleNetV2:
    """
    Constructs a ShuffleNetV2 with 1.5x output channels, as described in
    [ShuffleNet V2: Practical Guidelines for Efficient CNN Architecture Design](https://arxiv.org/abs/1807.11164).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _shufflenetv2([4, 8, 4], [24, 176, 352, 704, 1024], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["shufflenetv2_x1.5"])
    return model


def shufflenet_v2_x2_0(pretrained=False, **kwargs: Any) -> ShuffleNetV2:
    """
    Constructs a ShuffleNetV2 with 2.0x output channels, as described in
    [ShuffleNet V2: Practical Guidelines for Efficient CNN Architecture Design](https://arxiv.org/abs/1807.11164).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _shufflenetv2([4, 8, 4], [24, 244, 488, 976, 2048], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["shufflenetv2_x2.0"])
    return model



from typing import Optional, Tuple, Union

import equinox as eqx
import equinox.nn as nn
import jax
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array


class PatchEmbed(eqx.Module):
    """2D Image to Patch Embedding ported from Timm"""

    img_size: Tuple[int]
    patch_size: Tuple[int]
    grid_size: Tuple[int]
    num_patches: int
    flatten: bool
    proj: eqx.nn.Conv2d
    norm: eqx.Module

    def __init__(
        self,
        img_size: Union[int, Tuple[int]] = 224,
        patch_size: Union[int, Tuple[int]] = 16,
        in_chans: int = 3,
        embed_dim: int = 768,
        norm_layer: "eqx.Module" = None,
        flatten: bool = True,
        *,
        key: Optional["jax.random.PRNGKey"] = None,
    ):
        """
        **Arguments:**

        - `img_size`: The size of the input image. Defaults to `(224, 224)`
        - `patch_size`: Size of the patch to construct from the input image. Defaults to `(16, 16)`
        - `in_chans`: Number of input channels. Defaults to `3`
        - `embed_dim`: The dimension of the resulting embedding of the patch. Defaults to `768`
        - `norm_layer`: The normalisation to be applied on an input. Defaults to None
        - `flatten`: If enabled, the `2d` patches are flattened to `1d`
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        super().__init__()
        self.img_size = (
            img_size if isinstance(img_size, tuple) else (img_size, img_size)
        )
        self.patch_size = (
            patch_size if isinstance(patch_size, tuple) else (patch_size, patch_size)
        )
        self.grid_size = (
            self.img_size[0] // self.patch_size[0],
            self.img_size[1] // self.patch_size[1],
        )
        self.num_patches = self.grid_size[0] * self.grid_size[1]
        self.flatten = flatten
        if key is None:
            key = jrandom.PRNGKey(0)
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size, key=key
        )
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape`(in_chans, img_size[0], img_size[1])`.
        - `key`: Ignored
        """
        C, H, W = x.shape
        if H != self.img_size[0] or W != self.img_size[1]:
            raise ValueError(
                f"Input image height ({H},{W}) doesn't match model ({self.img_size})."
            )

        x = self.proj(x)
        if self.flatten:
            x = jax.vmap(jnp.ravel)(x)
            x = jnp.moveaxis(x, 0, -1)  # CHW -> NC
        x = self.norm(x)
        return x
